import re

from sae_lens.toolkit.pretrained_saes_directory import get_pretrained_saes_directory
from tabulate import tabulate
from tqdm.auto import tqdm


def all_loadable_saes() -> list[tuple[str, str, float, float]]:
    all_loadable_saes = []
    saes_directory = get_pretrained_saes_directory()
    for release, lookup in tqdm(saes_directory.items()):
        for sae_name in lookup.saes_map.keys():
            expected_var_explained = lookup.expected_var_explained[sae_name]
            expected_l0 = lookup.expected_l0[sae_name]
            all_loadable_saes.append(
                (release, sae_name, expected_var_explained, expected_l0)
            )

    return all_loadable_saes


def get_saes_from_regex(
    sae_regex_pattern: str, sae_id_pattern: str
) -> list[tuple[str, str]]:
    """
    Filter and retrieve SAEs based on regex patterns for release names and SAE IDs.

    This function searches through all loadable SAEs and returns those that match
    the provided regex patterns for both the release name and the SAE ID.

    Args:
        sae_regex_pattern (str): A regex pattern to match against SAE release names.
        sae_id_pattern (str): A regex pattern to match against SAE IDs.

    Returns:
        list[tuple[str, str]]: A list of tuples, where each tuple contains
        (release_name, sae_id) for SAEs matching both regex patterns.

    Example:
        >>> get_saes_from_regex(r"sae_bench_pythia.*", r"blocks\\.4\\.hook_resid_pre.*")
        [('sae_bench_pythia70m_sweep_standard_ctx128_0712', 'blocks.4.hook_resid_pre__trainer_0'),
         ('sae_bench_pythia70m_sweep_standard_ctx128_0712', 'blocks.4.hook_resid_pre__trainer_1'), ...]
    """
    sae_regex_compiled = re.compile(sae_regex_pattern)
    sae_id_compiled = re.compile(sae_id_pattern)
    all_saes = all_loadable_saes()
    filtered_saes = [
        sae
        for sae in all_saes
        if sae_regex_compiled.fullmatch(sae[0]) and sae_id_compiled.fullmatch(sae[1])
    ]

    # exclude the expected_var_explained and expected_l0 values
    filtered_saes = [(sae[0], sae[1]) for sae in filtered_saes]
    return filtered_saes


metadata_rows = [
    [data.model, data.release, data.repo_id, len(data.saes_map)]
    for data in get_pretrained_saes_directory().values()
]


# Print all SAE releases, sorted by base model
def print_all_sae_releases():
    """
    Print a table of all SAE releases, sorted by base model.
    """
    metadata_rows = [
        [data.model, data.release, data.repo_id, len(data.saes_map)]
        for data in get_pretrained_saes_directory().values()
    ]

    print(
        tabulate(
            sorted(metadata_rows, key=lambda x: x[0]),  # type: ignore
            headers=["model", "release", "repo_id", "n_saes"],
            tablefmt="simple_outline",
        )
    )


def print_release_details(release_name: str):
    """
    Print details of a specific SAE release.

    Args:
    release_name (str): The name of the release to display details for.
    """

    def format_value(value):
        if isinstance(value, dict):
            if not value:
                return "{}"
            return "{{{0!r}: {1!r}, ...}}".format(*next(iter(value.items())))  # noqa: UP030
        return repr(value)

    release = get_pretrained_saes_directory()[release_name]

    print(
        tabulate(
            [[k, format_value(v)] for k, v in release.__dict__.items()],
            headers=["Field", "Value"],
            tablefmt="simple_outline",
        )
    )


def select_saes_multiple_patterns(
    sae_regex_patterns: list[str],
    sae_block_patterns: list[str],
) -> list[tuple[str, str]]:
    assert len(sae_regex_patterns) == len(sae_block_patterns), "Length mismatch"

    selected_saes = []
    for sae_regex_pattern, sae_block_pattern in zip(
        sae_regex_patterns, sae_block_patterns
    ):
        selected_saes.extend(get_saes_from_regex(sae_regex_pattern, sae_block_pattern))
    assert len(selected_saes) > 0, "No SAEs selected"

    releases = set([release for release, _ in selected_saes])

    print(f"Selected SAEs from releases: {releases}")

    for release, sae in selected_saes:
        print(f"Sample SAEs: {release}, {sae}")

    return selected_saes



from pydantic import Field
from pydantic.dataclasses import dataclass

from sae_bench.evals.base_eval_output import BaseEvalConfig


@dataclass
class SparseProbingEvalConfig(BaseEvalConfig):
    random_seed: int = Field(
        default=42,
        title="Random Seed",
        description="Random seed",
    )

    dataset_names: list[str] = Field(
        default_factory=lambda: [
            "LabHC/bias_in_bios_class_set1",
            "LabHC/bias_in_bios_class_set2",
            "LabHC/bias_in_bios_class_set3",
            "canrager/amazon_reviews_mcauley_1and5",
            "canrager/amazon_reviews_mcauley_1and5_sentiment",
            "codeparrot/github-code",
            "fancyzhx/ag_news",
            "Helsinki-NLP/europarl",
        ],
        title="Dataset Names",
        description="List of dataset names. We have at most 5 class names in a single subset, which is why we have multiple bias_in_bios class subsets.",
    )

    probe_train_set_size: int = Field(
        default=4000,
        title="Probe Train Set Size",
        description="Probe train set size",
    )
    probe_test_set_size: int = Field(
        default=1000,
        title="Probe Test Set Size",
        description="Probe test set size",
    )
    context_length: int = Field(
        default=128,
        title="LLM Context Length",
        description="The maximum length of each input to the LLM. Any longer inputs will be truncated, keeping only the beginning.",
    )

    sae_batch_size: int = Field(
        default=125,
        title="SAE Batch Size",
        description="SAE batch size, inference only",
    )
    llm_batch_size: int | None = Field(
        default=None,
        title="LLM Batch Size",
        description="LLM batch size. This is set by default in the main script, or it can be set with a command line argument.",
    )  # type: ignore
    llm_dtype: str = Field(
        default="",
        title="LLM Data Type",
        description="LLM data type. This is set by default in the main script, or it can be set with a command line argument.",
    )

    model_name: str = Field(
        default="",
        title="Model Name",
        description="Model name. Must be set with a command line argument.",
    )

    k_values: list[int] = Field(
        default_factory=lambda: [1, 2, 5],
        title="K Values",
        description="K represents the number of SAE features or residual stream channels we train the linear probe on. We iterate over all values of K.",
    )

    lower_vram_usage: bool = Field(
        default=False,
        title="Lower Memory Usage",
        description="Lower GPU memory usage by doing more computation on the CPU. Useful on 1M width SAEs. Will be slower and require more system memory.",
    )



from pydantic import ConfigDict, Field
from pydantic.dataclasses import dataclass

from sae_bench.evals.base_eval_output import (
    DEFAULT_DISPLAY,
    BaseEvalOutput,
    BaseMetricCategories,
    BaseMetrics,
    BaseResultDetail,
)
from sae_bench.evals.ravel.eval_config import RAVELEvalConfig

EVAL_TYPE_ID_RAVEL = "ravel"


@dataclass
class RAVELMetricResults(BaseMetrics):
    disentanglement_score: float = Field(
        title="Disentanglement Score",
        description="Mean of cause and isolation scores across RAVEL datasets.",
        json_schema_extra=DEFAULT_DISPLAY,
    )
    cause_score: float = Field(
        title="Cause Score",
        description="Cause score: Patching attribute-related SAE latents. High cause accuracy indicates that the SAE latents are related to the attribute.",
        json_schema_extra=DEFAULT_DISPLAY,
    )
    isolation_score: float = Field(
        title="Isolation Score",
        description="Isolation score: Patching SAE latents related to another attribute. High isolation accuracy indicates that latents related to another attribute are not related to this attribute.",
        json_schema_extra=DEFAULT_DISPLAY,
    )


@dataclass
class RAVELMetricCategories(BaseMetricCategories):
    ravel: RAVELMetricResults = Field(
        title="RAVEL",
        description="RAVEL metrics",
        json_schema_extra=DEFAULT_DISPLAY,
    )


@dataclass(config=ConfigDict(title="RAVEL"))
class RAVELEvalOutput(
    BaseEvalOutput[RAVELEvalConfig, RAVELMetricCategories, BaseResultDetail]
):
    # This will end up being the description of the eval in the UI.
    """
    An evaluation using SAEs for targeted modification of language model output. We leverage the RAVEL dataset of entity-attribute pairs. After filtering for known pairs, we identify attribute-related SAE latents and deterimine the effect on model predictions with activation patching experiments.
    """

    eval_config: RAVELEvalConfig
    eval_id: str
    datetime_epoch_millis: int
    eval_result_metrics: RAVELMetricCategories
    eval_type_id: str = Field(default=EVAL_TYPE_ID_RAVEL)



from dataclasses import dataclass

import numpy as np
import torch
from sae_lens import SAE
from tqdm.autonotebook import tqdm
from transformer_lens import HookedTransformer

from sae_bench.evals.absorption.prompting import (
    Formatter,
    SpellingPrompt,
    create_icl_prompt,
    first_letter_formatter,
)
from sae_bench.evals.absorption.util import batchify

EPS = 1e-8


@dataclass
class FeatureScore:
    feature_id: int
    activation: float
    probe_cos_sim: float

    @property
    def probe_projection(self) -> float:
        return self.activation * self.probe_cos_sim


@dataclass
class WordAbsorptionResult:
    word: str
    prompt: str
    probe_projection: float
    main_feature_scores: list[FeatureScore]
    top_projection_feature_scores: list[FeatureScore]
    absorption_fraction: float
    is_full_absorption: bool


@dataclass
class AbsorptionResults:
    main_feature_ids: list[int]
    word_results: list[WordAbsorptionResult]


@dataclass
class FeatureAbsorptionCalculator:
    """
    Feature absorption calculator for spelling tasks.

    Absorption is defined by the following criteria:
    - The main features for a concept do not fire
    - The top feature is aligned with a probe trained on that concept
    - The top feature contributes a significant portion of the total activation probe projection
    """

    model: HookedTransformer
    icl_word_list: list[str]
    max_icl_examples: int | None = None
    base_template: str = "{word}:"
    answer_formatter: Formatter = first_letter_formatter()
    example_separator: str = "\n"
    shuffle_examples: bool = True
    # Mistral tokenizer handles the first token incorrectly if we don't do this
    prepend_separator_to_first_example: bool = True
    # the position to read activations from (depends on the template)
    word_token_pos: int = -2
    batch_size: int = 10
    topk_feats: int = 10

    # the cosine similarity between the top projecting feature and the probe must be at least this high to count as absorption (full absorption only)
    full_absorption_probe_cos_sim_threshold: float = 0.025
    # the cosine similarity between each potential absorbing latent and the probe must be at least this high to count as absorption (absorption fraction only)
    absorption_fraction_probe_cos_sim_threshold: float = 0.1
    # the total probe projection of the potential absorbing latents must contribute at least this much to the probe projection to count as absorption (both absorption metrics)
    probe_projection_proportion_threshold: float = 0.4
    # the maximum number of latents that can be considered to collectively compensate for the reduced activation of a potentially absorbed latent (absorption fraction only)
    absorption_fraction_max_absorbing_latents: int = 3

    def _build_prompts(self, words: list[str]) -> list[SpellingPrompt]:
        return [
            create_icl_prompt(
                word,
                examples=self.icl_word_list,
                base_template=self.base_template,
                answer_formatter=self.answer_formatter,
                example_separator=self.example_separator,
                max_icl_examples=self.max_icl_examples,
                shuffle_examples=self.shuffle_examples,
                prepend_separator_to_first_example=self.prepend_separator_to_first_example,
            )
            for word in words
        ]

    def _is_full_absorption(
        self,
        probe_projection: float,
        main_feature_scores: list[FeatureScore],
        top_projection_feature_scores: list[FeatureScore],
    ) -> bool:
        # if any of the main features fired, this isn't absorption
        if not all(score.activation < EPS for score in main_feature_scores):
            return False
        # If the top firing feature isn't aligned with the probe, this isn't absorption
        if (
            top_projection_feature_scores[0].probe_cos_sim
            < self.full_absorption_probe_cos_sim_threshold
        ):
            return False
        # If the probe isn't even activated, this can't be absorption
        if probe_projection < 0:
            return False
        # If the top firing feature doesn't contribute much to the total probe projection, this isn't absorption
        proj_proportion = (
            top_projection_feature_scores[0].probe_projection / probe_projection
        )
        if proj_proportion < self.probe_projection_proportion_threshold:
            return False
        return True

    @torch.inference_mode()
    def calculate_absorption(
        self,
        sae: SAE,
        words: list[str],
        probe_direction: torch.Tensor,
        main_feature_ids: list[int],
        layer: int,
        show_progress: bool = True,
    ) -> AbsorptionResults:
        """
        This method calculates the absorption for each word in the list of words
        """
        if probe_direction.ndim != 1:
            raise ValueError("probe_direction must be 1D")
        # make sure the probe direction is a unit vector
        probe_direction = probe_direction / probe_direction.norm()
        prompts = self._build_prompts(words)
        self._validate_prompts_are_same_length(prompts)
        results: list[WordAbsorptionResult] = []
        cos_sims = (
            torch.nn.functional.cosine_similarity(
                probe_direction.to(sae.device), sae.W_dec, dim=-1
            )
            .float()
            .cpu()
        )
        hook_point = f"blocks.{layer}.hook_resid_post"
        for batch_prompts in batchify(prompts, batch_size=self.batch_size):
            batch_acts = self.model.run_with_cache(
                [p.base for p in batch_prompts],
                names_filter=[hook_point],
            )[1][hook_point][:, self.word_token_pos, :]
            batch_sae_acts = sae.encode(batch_acts)
            batch_sae_probe_projections = batch_sae_acts * cos_sims.to(
                batch_sae_acts.device
            )
            batch_probe_projections = batch_acts @ probe_direction.to(
                device=batch_sae_acts.device, dtype=batch_sae_acts.dtype
            )
            for i, prompt in enumerate(tqdm(batch_prompts, disable=not show_progress)):
                sae_acts = batch_sae_acts[i]
                act_probe_proj = batch_probe_projections[i].cpu().item()
                sae_act_probe_proj = batch_sae_probe_projections[i].cpu()

                ### calculate absorption_fraction ###

                # GT probe proj of main feats
                main_feats_probe_proj = (
                    torch.sum(sae_act_probe_proj[main_feature_ids]).cpu().item()
                )

                # GT probe proj of other feats
                potential_absorbers_mask = torch.ones(
                    sae_act_probe_proj.size(0), dtype=torch.bool
                )
                potential_absorbers_mask[main_feature_ids] = False
                potential_absorbers_mask &= (
                    cos_sims >= self.absorption_fraction_probe_cos_sim_threshold
                )
                potential_absorbers_mask &= sae_act_probe_proj > 0
                potential_absorbers_probe_proj = sae_act_probe_proj[
                    potential_absorbers_mask
                ]
                top_potential_absorbers_probe_proj = (
                    potential_absorbers_probe_proj.topk(
                        k=min(
                            self.absorption_fraction_max_absorbing_latents,
                            potential_absorbers_probe_proj.numel(),
                        )
                    ).values
                )
                top_potential_absorbers_total_probe_proj = (
                    torch.sum(top_potential_absorbers_probe_proj).cpu().item()
                )

                # final absorption_fraction calculation
                top_potential_absorbers_probe_proj_proportion = (
                    top_potential_absorbers_total_probe_proj / act_probe_proj
                )
                if (
                    main_feats_probe_proj >= act_probe_proj
                    or top_potential_absorbers_probe_proj_proportion
                    < self.probe_projection_proportion_threshold
                ):
                    absorption_fraction = 0.0
                elif main_feats_probe_proj <= 0.0:
                    absorption_fraction = 1.0
                else:
                    unaccounted_probe_proj = act_probe_proj - main_feats_probe_proj
                    absorption_probe_proj = min(
                        top_potential_absorbers_total_probe_proj, unaccounted_probe_proj
                    )
                    absorption_fraction = absorption_probe_proj / (
                        absorption_probe_proj + main_feats_probe_proj
                    )
                    absorption_fraction = np.clip(absorption_fraction, 0.0, 1.0)

                ### determine whether this is full absorption with a single absorbing latent ###

                with torch.inference_mode():
                    # sort by negative ig score
                    top_proj_feats = sae_act_probe_proj.topk(
                        self.topk_feats
                    ).indices.tolist()
                    main_feature_scores = _get_feature_scores(
                        main_feature_ids,
                        probe_cos_sims=cos_sims,
                        sae_acts=sae_acts,
                    )
                    top_projection_feature_scores = _get_feature_scores(
                        top_proj_feats,
                        probe_cos_sims=cos_sims,
                        sae_acts=sae_acts,
                    )
                    is_full_absorption = self._is_full_absorption(
                        probe_projection=act_probe_proj,
                        top_projection_feature_scores=top_projection_feature_scores,
                        main_feature_scores=main_feature_scores,
                    )

                results.append(
                    WordAbsorptionResult(
                        word=prompt.word,
                        prompt=prompt.base,
                        probe_projection=act_probe_proj,
                        main_feature_scores=main_feature_scores,
                        top_projection_feature_scores=top_projection_feature_scores,
                        absorption_fraction=absorption_fraction,
                        is_full_absorption=is_full_absorption,
                    )
                )
        return AbsorptionResults(
            main_feature_ids=main_feature_ids,
            word_results=results,
        )

    def _validate_prompts_are_same_length(self, prompts: list[SpellingPrompt]):
        "Validate that all prompts have the same token length"
        token_lens = {len(self.model.to_tokens(p.base)[0]) for p in prompts}
        if len(token_lens) > 1:
            raise ValueError(
                "All prompts must have the same token length! Variable-length prompts are not yet supported."
            )


def _get_feature_scores(
    feature_ids: list[int],
    probe_cos_sims: torch.Tensor,
    sae_acts: torch.Tensor,
) -> list[FeatureScore]:
    return [
        FeatureScore(
            feature_id=feature_id,
            probe_cos_sim=probe_cos_sims[feature_id].item(),
            activation=sae_acts[feature_id].item(),
        )
        for feature_id in feature_ids
    ]



import json
import sys
from pathlib import Path

from sae_bench.evals.core.eval_output import CoreEvalOutput
from sae_bench.evals.core.main import convert_feature_metrics

# This script is used to convert an old-format eval output to the new format.
# The old format is no longer produced, so you don't need to use this script.

# load input directory from command line
input_dir = Path(sys.argv[1])

# Get all JSON files in directory, sorted alphabetically
input_files = sorted(input_dir.glob("*.json"))

if not input_files:
    print(f"No JSON files found in {input_dir}")
    sys.exit(1)

# Create outputs directory if it doesn't exist
output_dir = input_dir / "converted_outputs"
output_dir.mkdir(exist_ok=True)

# Convert each file
for input_file in input_files:
    print(f"Converting {input_file}")
    output_file = output_dir / input_file.name
    with open(input_file) as f:
        data = json.load(f)
    feature_metrics = convert_feature_metrics(data["eval_result_details"][0])
    data["eval_result_details"] = feature_metrics
    with open(output_file, "w") as f:
        eval_output = CoreEvalOutput(
            eval_config=data["eval_config"],
            eval_id=data["eval_id"],
            datetime_epoch_millis=data["datetime_epoch_millis"],
            eval_result_metrics=data["eval_result_metrics"],
            eval_result_details=data["eval_result_details"],
            eval_result_unstructured=data.get("eval_result_unstructured", {}),
            sae_bench_commit_hash=data["sae_bench_commit_hash"],
            sae_lens_id=data["sae_lens_id"],
            sae_lens_release_id=data["sae_lens_release_id"],
            sae_lens_version=data["sae_lens_version"],
            sae_cfg_dict=data["sae_cfg_dict"],
        )
        eval_output.to_json_file(str(output_file))



"""
RAVEL Entity Prompt Data Module

This module provides functionality for handling and processing entity prompt data
for the RAVEL evaluation benchmark.
"""

import json
import os
import random
from copy import deepcopy
from dataclasses import dataclass

from huggingface_hub import snapshot_download
from tqdm import tqdm
from transformers import (
    AutoModelForCausalLM,
    AutoTokenizer,
    PreTrainedModel,
    PreTrainedTokenizer,
    PreTrainedTokenizerFast,
)

from sae_bench.evals.ravel.eval_config import RAVELEvalConfig
from sae_bench.evals.ravel.generation import generate_batched
from sae_bench.evals.ravel.validation import evaluate_completion


@dataclass
class AttributePrompt:
    """Represents an attribute_type with its associated prompt templates."""

    attribute_type: str
    templates: list[str]


@dataclass
class Prompt:
    """Represents a single prompt with its associated data."""

    text: str  # Template with inserted entity label.
    template: str  # The template string with %s placeholder for entity label.
    attribute_type: str  # The abstract attribute type, eg. "Country".
    attribute_label: str  # The concrete attribute label, eg. "Finland".
    entity_label: str  # The entity label, eg. "Helsinki".
    context_split: str  # The context split, "train"/"val".
    entity_split: str  # The entity split, "train"/"val".
    input_ids: list[int] | None = None  # Tokenized text.
    final_entity_token_pos: int | None = (
        None  # Position of the final entity token in the input_ids, as counted from the end (negative index)
    )
    attention_mask: list[int] | None = None
    attribute_generation: str | None = (
        None  # Given the text, the generated next tokens which may contain the attribute label, decoded to string.
    )
    first_generated_token_id: int | None = (
        None  # The first generated token id from attribute_generation.
    )
    is_correct: bool | None = (
        None  # Whether the attribute generation contains the attribute label.
    )


def get_instance_name(
    entity_type: str,
    model_name: str,
    downsample: int | None = None,
    top_n_entities: int | None = None,
) -> str:
    model_name_str = model_name.replace("/", "--")
    instance_name = f"{entity_type}_{model_name_str}_downsampled-{downsample}"
    if top_n_entities:
        instance_name += f"_top-{top_n_entities}-entities_filtered_dataset.json"
    return instance_name


class RAVELInstance:
    """
    The dataset for the RAVEL Benchmark is created in two steps:
    1. Create a RAVELInstance object from the raw RAVEL dataset files. This will contain all (num_templates x num_entities) prompts.
        We'll have to check whether the model correctly answers these prompts, and only want to keep the entities with the most correctly answered prompts.
        Therefore, we'll have to generate completions and evaluate correctness for all prompts, once for each model.
        Optionally, we can downsample the dataset before generating completions. This risks loosing entities with low coverage, but is faster.
        This can take a while, so we'll save the RAVELInstance object json after each model and host on huggingface.
    2. Create a RAVELFilteredDataset object from the RAVELInstance object. This will contain a filtered subset of the prompts, padded to the max prompt length.
        Filtering:
        - Only keep entities with the most correctly answered prompts.
        - Only keep templates with the most correctly answered prompts.
        - Pad prompts to the max prompt length.
        - Save as json.
    """

    def __init__(self):
        self.prompts = []  # list of Prompt objects
        self.entityLBL_attrTYP_attrLBL = {}  # entity label -> attribute type -> attribute label
        self.template_splits = {}  # template -> 'train'/'val'
        self.entity_splits = {}  # entity -> 'train'/'val'
        self.attribute_type_to_templates = {}  # attribute type -> (templates x entities) Prompt objects
        self.config = {}

        # If this exists, we only tokenize prompts for these attribute types.
        self.attribute_types: list[str] | None = None

    @classmethod
    def create_from_files(
        cls,
        config: RAVELEvalConfig,
        entity_type: str,
        data_dir: str,
        tokenizer: PreTrainedTokenizer | PreTrainedTokenizerFast,
        model: PreTrainedModel,
        model_name: str,
        attribute_types: list[str] | None = None,
        downsample: int | None = None,
    ) -> "RAVELInstance":
        instance = cls()
        instance.attribute_types = attribute_types

        instance.initialize_config(entity_type, model_name, downsample)
        save_path = os.path.join(
            config.artifact_dir,
            f"{instance.config['instance_name']}_unfiltered_full_instance.json",
        )

        if os.path.exists(save_path):
            print(f"Loading instance from {save_path}.")
            return instance.load(save_path)

        print("Loading files.")
        instance.load_files(entity_type, data_dir)

        print("Tokenizing prompts.")
        instance.build_and_tokenize_prompts(tokenizer)

        # Optional: Downsample to fewer prompts.
        if downsample:
            print(f"Downsample to {downsample} prompts.")
            instance.downsample_(downsample)

        print("Generate completions.")
        instance.generate_completions(
            model,
            tokenizer,
            max_new_tokens=config.n_generated_tokens,
            llm_batch_size=config.llm_batch_size,
        )

        print("Evaluate correctness.")
        instance.evaluate_correctness()

        print("Filter correct completions.")
        instance.filter_correct_()

        print("Save filtered dataset.")
        instance.save_as_instance(save_path)
        return instance

    def initialize_config(
        self, entity_type: str, model_name: str, downsample: int | None = None
    ):
        instance_name = get_instance_name(entity_type, model_name, downsample)
        self.config = {
            "entity_type": entity_type,
            "model_name": model_name,
            "downsample": downsample,
            "instance_name": instance_name,
        }
        return self.config

    def build_and_tokenize_prompts(
        self,
        tokenizer: PreTrainedTokenizer | PreTrainedTokenizerFast,
    ) -> list[Prompt]:
        """
        Load the full RAVEL dataset from files.
        """
        # Tokenize prompts from (template x entity) combinations.
        for entity_label in tqdm(
            self.entityLBL_attrTYP_attrLBL,
            total=len(self.entityLBL_attrTYP_attrLBL),
            desc="Tokenizing prompts",
        ):
            for attribute_type, templates in self.attribute_type_to_templates.items():
                if self.attribute_types and attribute_type not in self.attribute_types:
                    continue
                for template in templates:
                    text = template % entity_label
                    encoded = tokenizer.encode(text)
                    if isinstance(
                        encoded[0], list
                    ):  # TODO: actually check this and remove this check.
                        raise ValueError(
                            "Batch dimension not supported. Please adapt tokenization"
                        )

                    remainder = template.split("%s")[1]
                    encoded_remainder = tokenizer.encode(remainder)
                    final_pos = -len(encoded_remainder)

                    self.prompts.append(
                        Prompt(
                            text=text,
                            template=template,
                            attribute_type=attribute_type,
                            attribute_label=self.entityLBL_attrTYP_attrLBL[
                                entity_label
                            ][attribute_type],
                            entity_label=entity_label,
                            context_split=self.template_splits[template],
                            entity_split=self.entity_splits[entity_label],
                            input_ids=encoded,
                            final_entity_token_pos=final_pos,
                        )
                    )
        return self.prompts

    def load_files(
        self,
        entity_type: str,
        data_dir: str,
    ) -> None:
        # Define file paths and names
        base_dir = os.path.join(data_dir, "base")
        os.makedirs(base_dir, exist_ok=True)

        required_files = [
            f"ravel_{entity_type}_attribute_to_prompts.json",
            f"ravel_{entity_type}_prompt_to_split.json",
            f"ravel_{entity_type}_entity_attributes.json",
            f"ravel_{entity_type}_entity_to_split.json",
        ]

        # Check if any file is missing
        if any(not os.path.exists(os.path.join(base_dir, f)) for f in required_files):
            print("Downloading RAVEL dataset from HuggingFace...")
            snapshot_download(
                repo_id="adamkarvonen/ravel_prompts",
                repo_type="dataset",
                local_dir=base_dir,
                local_dir_use_symlinks=False,
                allow_patterns="*.json",
            )

        # Load data files
        with open(
            os.path.join(
                data_dir, "base", f"ravel_{entity_type}_entity_attributes.json"
            )
        ) as f:
            self.entityLBL_attrTYP_attrLBL = json.load(f)
        with open(
            os.path.join(data_dir, "base", f"ravel_{entity_type}_prompt_to_split.json")
        ) as f:
            self.template_splits = json.load(f)
        with open(
            os.path.join(data_dir, "base", f"ravel_{entity_type}_entity_to_split.json")
        ) as f:
            self.entity_splits = json.load(f)
        with open(
            os.path.join(
                data_dir, "base", f"ravel_{entity_type}_attribute_to_prompts.json"
            )
        ) as f:
            self.attribute_type_to_templates = json.load(f)

    def downsample_(self, n: int) -> None:
        sampled_keys = random.sample(list(range(len(self.prompts))), n)
        sampled_prompts = [self.prompts[k] for k in sampled_keys]
        self._filter_data_(sampled_prompts)

    def __len__(self) -> int:
        return len(self.prompts)

    def get_prompts_by_split(self, context_split: str) -> list[Prompt]:
        """Return all prompts with the given context split."""
        return [
            prompt for prompt in self.prompts if prompt.context_split == context_split
        ]

    def get_entities(self, split: str | None = None) -> list[str]:
        """Return all entities with the given split."""
        if split is None:
            return list(self.entity_splits.keys())
        return [
            entity_label
            for entity_label, entity_split in self.entity_splits.items()
            if entity_split == split
        ]

    def get_attributes(self) -> list[str]:
        """Return all attribute types."""
        return list(self.attribute_type_to_templates.keys())

    def get_prompt_by_text(self, text: str) -> Prompt | None:
        """Return the unique prompt with the given text, if available."""
        return next((p for p in self.prompts if p.text == text), None)

    def get_prompts_by_template(self, template: str) -> list[Prompt]:
        """Return all prompts with the given template."""
        return [p for p in self.prompts if p.template == template]

    def get_prompts_by_attribute(
        self, attribute: str, n_samples: int | None = None
    ) -> list[Prompt]:
        """Return all prompts with the given attribute type."""
        prompts = [p for p in self.prompts if p.attribute_type == attribute]
        if n_samples:
            if n_samples > len(prompts):
                print(
                    f"Warning: Requested {n_samples} samples but only {len(prompts)} available"
                )
            return prompts[:n_samples]
        return prompts

    def get_prompts_by_entity(self, entity_label: str) -> list[Prompt]:
        """Return all prompts with the given entity label."""
        return [p for p in self.prompts if p.entity_label == entity_label]

    def generate_completions(
        self,
        model: PreTrainedModel,
        tokenizer: PreTrainedTokenizer | PreTrainedTokenizerFast,
        max_new_tokens: int,
        llm_batch_size: int = 32,
        **kwargs,
    ) -> None:
        """Generate completions for all prompts."""

        token_ids = [p.input_ids for p in self.prompts]
        attention_masks = None  # Attention masks are computed per batch dependent on padding within generate_batched.
        completions, first_token_ids = (
            generate_batched(  # TODO: Add tokenization to this function.
                model,
                tokenizer,
                input_ids_BL=token_ids,  # type: ignore
                attention_mask_BL=attention_masks,
                max_new_tokens=max_new_tokens,
                llm_batch_size=llm_batch_size,
                return_first_generated_token=True,
                **kwargs,
            )
        )

        for prompt, completion, first_token_id in zip(
            self.prompts, completions, first_token_ids
        ):
            prompt.attribute_generation = completion
            prompt.first_generated_token_id = first_token_id  # type: ignore

    def _filter_data_(self, filtered_prompts: list[Prompt]) -> None:
        """Filter the data based on the filtered prompts."""
        filtered_entity_labels = set(p.entity_label for p in filtered_prompts)
        filtered_attribute_types = set(p.attribute_type for p in filtered_prompts)
        filtered_templates = set(p.template for p in filtered_prompts)

        filtered_entity_label_to_attribute_type = {
            e: attrTYP_attrLBL
            for e, attrTYP_attrLBL in self.entityLBL_attrTYP_attrLBL.items()
            if e
            in filtered_entity_labels  # NOTE attributes listed do not necessarily have a prompt the model can answer correctly.
        }
        filtered_template_splits = {
            t: split
            for t, split in self.template_splits.items()
            if t in filtered_templates
        }
        filtered_entity_splits = {
            e: split
            for e, split in self.entity_splits.items()
            if e in filtered_entity_labels
        }
        filtered_attribute_prompts = {
            attribute_type: [t for t in templates if t in filtered_templates]
            for attribute_type, templates in self.attribute_type_to_templates.items()
            if attribute_type in filtered_attribute_types
        }

        # Update the instance attributes.
        self.prompts = filtered_prompts
        self.entityLBL_attrTYP_attrLBL = filtered_entity_label_to_attribute_type
        self.template_splits = filtered_template_splits
        self.entity_splits = filtered_entity_splits
        self.attribute_type_to_templates = filtered_attribute_prompts

    def filter_correct_(self):
        correct_prompts = [p for p in self.prompts if p.is_correct]
        self._filter_data_(correct_prompts)

    def evaluate_correctness(self):
        """Evaluate whether the generated completion contains the expected attribute label."""
        for prompt in self.prompts:
            if prompt.attribute_generation is not None:
                prompt.is_correct = evaluate_completion(
                    text=prompt.text,
                    expected_label=prompt.attribute_label,
                    completion=prompt.attribute_generation,
                )

    def get_accuracy_stats(self):
        """Get accuracy stats for all prompts."""
        stats = {}
        for prompt in self.prompts:
            if prompt.is_correct is not None:
                key = (prompt.entity_label, prompt.template)
                if key not in stats:
                    stats[key] = {"correct": 0, "total": 0}
                stats[key]["total"] += 1
                if prompt.is_correct:
                    stats[key]["correct"] += 1
        return stats

    def calculate_average_accuracy(self):
        """Calculate the average accuracy of the model."""
        correct = sum(1 for p in self.prompts if p.is_correct)
        total = len(self.prompts)
        return correct / total if total > 0 else 0

    def filter_prompts_by_template_format(self):
        return {
            text: p
            for text, p in self.prompts.items()  # type: ignore
            if p.template.count("%s") == 1
        }

    def filter_top_entities(self, top_n_entities=400):
        stats = self.get_accuracy_stats()

        # Get top entities
        entity_scores = {}
        for (entity, _), stat in stats.items():
            entity_scores[entity] = entity_scores.get(entity, 0) + stat["correct"]
        kept_entities = set(
            sorted(entity_scores, key=lambda x: entity_scores[x], reverse=True)[
                :top_n_entities
            ]
        )

        filtered_prompts = [p for p in self.prompts if p.entity_label in kept_entities]
        return self._filter_data_(filtered_prompts)

    def save_as_instance(self, save_path: str):
        """Save the RAVELInstance object to a json file."""
        ravel_instance_dict = {
            "prompts": [p.__dict__ for p in self.prompts],
            "entityLBL_attrTYP_attrLBL": self.entityLBL_attrTYP_attrLBL,
            "template_splits": self.template_splits,
            "entity_splits": self.entity_splits,
            "attribute_type_to_templates": self.attribute_type_to_templates,
            "config": self.config,
        }
        with open(save_path, "w") as f:
            json.dump(ravel_instance_dict, f)
        return ravel_instance_dict

    @classmethod
    def load(cls, load_path: str):
        """Load the RAVELInstance object from a json file."""
        with open(load_path) as f:
            ravel_instance_dict = json.load(f)
        fresh_instance = cls()
        fresh_instance.prompts = [Prompt(**p) for p in ravel_instance_dict["prompts"]]
        fresh_instance.entityLBL_attrTYP_attrLBL = ravel_instance_dict[
            "entityLBL_attrTYP_attrLBL"
        ]
        fresh_instance.template_splits = ravel_instance_dict["template_splits"]
        fresh_instance.entity_splits = ravel_instance_dict["entity_splits"]
        fresh_instance.attribute_type_to_templates = ravel_instance_dict[
            "attribute_type_to_templates"
        ]
        fresh_instance.config = ravel_instance_dict["config"]
        return fresh_instance

    def create_and_save_filtered_dataset(
        self,
        artifact_dir: str,
        top_n_entities: int,
    ) -> "RAVELFilteredDataset":
        """Create and save the filtered dataset."""
        self.filter_top_entities(top_n_entities)

        config = deepcopy(self.config)
        config["top_n_entities"] = top_n_entities
        config["instance_name"] = (
            config["instance_name"] + f"_top-{top_n_entities}-entities"
        )
        prompt_dict = {
            "prompts": [p.__dict__ for p in self.prompts],
            "config": config,
        }

        filtered_dataset_path = os.path.join(
            artifact_dir, f"{config['instance_name']}_filtered_dataset.json"
        )
        with open(filtered_dataset_path, "w") as f:
            json.dump(prompt_dict, f)

        return RAVELFilteredDataset.from_dict(prompt_dict)


class RAVELFilteredDataset:
    def __init__(self, prompts: list[Prompt], config: dict):
        self.prompts = prompts
        self.config = config

    def get_prompts_by_attribute(self, attribute: str) -> list[Prompt]:
        return [p for p in self.prompts if p.attribute_type == attribute]

    def get_prompts_by_entity(self, entity: str) -> list[Prompt]:
        return [p for p in self.prompts if p.entity_label == entity]

    def get_prompts_by_template(self, template: str) -> list[Prompt]:
        return [p for p in self.prompts if p.template == template]

    def get_prompts_by_context_split(self, split: str) -> list[Prompt]:
        return [p for p in self.prompts if p.context_split == split]

    def get_prompts_by_entity_split(self, split: str) -> list[Prompt]:
        return [p for p in self.prompts if p.entity_split == split]

    def __len__(self):
        return len(self.prompts)

    def __getitem__(self, idx):
        return self.prompts[idx]

    def __iter__(self):
        return iter(self.prompts)

    def __contains__(self, item):
        return item in self.prompts

    def __repr__(self):
        return f"RAVELFilteredDataset(prompts={self.prompts})"

    def __str__(self):
        return f"RAVELFilteredDataset(prompts={self.prompts})"

    def save(self, artifact_dir: str):
        prompt_dict = {
            "prompts": [p.__dict__ for p in self.prompts],
            "config": self.config,
        }
        save_path = os.path.join(
            artifact_dir,
            f"{self.config['instance_name']}_top-{self.config['top_n_entities']}-entities_filtered_dataset.json",
        )
        with open(save_path, "w") as f:
            json.dump(prompt_dict, f)

    @classmethod
    def from_dict(cls, prompt_dict: dict):
        return cls(
            prompts=[Prompt(**p) for p in prompt_dict["prompts"]],
            config=prompt_dict["config"],
        )

    @classmethod
    def load(cls, load_path: str):
        with open(load_path) as f:
            prompt_dict = json.load(f)
        return cls.from_dict(prompt_dict)


if __name__ == "__main__":
    import sae_bench.sae_bench_utils.general_utils as general_utils

    # Load model and tokenizer
    config = RAVELEvalConfig()
    device = "cuda:0"

    LLM_NAME_MAP = {
        "gemma-2-2b": "google/gemma-2-2b",
    }
    config.model_name = LLM_NAME_MAP[config.model_name]
    llm_dtype = general_utils.str_to_dtype(config.llm_dtype)
    config.llm_batch_size = 32
    config.full_dataset_downsample = None

    model = AutoModelForCausalLM.from_pretrained(
        config.model_name,
        device_map=device,
        torch_dtype=llm_dtype,
        attn_implementation="eager",
    )
    tokenizer = AutoTokenizer.from_pretrained(config.model_name)

    # Create full RAVELInstance, no downsample, generate completions, filter for correct completions, save.
    entity_type = list(config.entity_attribute_selection.keys())[0]
    attribute_types = config.entity_attribute_selection[entity_type]
    print("Loading and tokenizing full dataset")
    full_dataset = RAVELInstance.create_from_files(
        config=config,
        entity_type=entity_type,
        tokenizer=tokenizer,
        data_dir=config.artifact_dir,
        model=model,
        model_name=config.model_name,
        attribute_types=attribute_types,
        downsample=config.full_dataset_downsample,
    )

    # Test loading the full dataset.
    instance_filename = (
        full_dataset.config["instance_name"] + "_unfiltered_full_instance.json"
    )
    instance_path = os.path.join(config.artifact_dir, instance_filename)
    full_dataset = RAVELInstance.load(instance_path)

    # Create filtered dataset.
    filtered_dataset = full_dataset.create_and_save_filtered_dataset(
        artifact_dir=config.artifact_dir,
        top_n_entities=config.top_n_entities,
    )

    # Test loading the filtered dataset.
    filtered_dataset_filename = (
        filtered_dataset.config["instance_name"] + "_filtered_dataset.json"
    )
    filtered_dataset_path = os.path.join(config.artifact_dir, filtered_dataset_filename)
    filtered_dataset = RAVELFilteredDataset.load(filtered_dataset_path)



import argparse
import gc
import os
import pickle
import random
import shutil
import time
from dataclasses import asdict
from datetime import datetime

import einops
import torch
from sae_lens import SAE
from tqdm import tqdm
from transformer_lens import HookedTransformer

import sae_bench.evals.scr_and_tpp.dataset_creation as dataset_creation
import sae_bench.evals.sparse_probing.probe_training as probe_training
import sae_bench.sae_bench_utils.activation_collection as activation_collection
import sae_bench.sae_bench_utils.dataset_info as dataset_info
import sae_bench.sae_bench_utils.dataset_utils as dataset_utils
import sae_bench.sae_bench_utils.general_utils as general_utils
from sae_bench.evals.scr_and_tpp.eval_config import ScrAndTppEvalConfig
from sae_bench.evals.scr_and_tpp.eval_output import (
    EVAL_TYPE_ID_SCR,
    EVAL_TYPE_ID_TPP,
    ScrEvalOutput,
    ScrMetricCategories,
    ScrMetrics,
    ScrResultDetail,
    TppEvalOutput,
    TppMetricCategories,
    TppMetrics,
    TppResultDetail,
)
from sae_bench.sae_bench_utils import (
    get_eval_uuid,
    get_sae_bench_version,
    get_sae_lens_version,
)
from sae_bench.sae_bench_utils.sae_selection_utils import get_saes_from_regex

COLUMN2_VALS_LOOKUP = {
    "LabHC/bias_in_bios_class_set1": ("male", "female"),
    "canrager/amazon_reviews_mcauley_1and5": (1.0, 5.0),
}


@torch.no_grad()
def get_effects_per_class_precomputed_acts(
    sae: SAE,
    probe: probe_training.Probe,
    class_idx: str,
    precomputed_acts: dict[str, torch.Tensor],
    perform_scr: bool,
    sae_batch_size: int,
) -> torch.Tensor:
    inputs_train_BLD, labels_train_B = probe_training.prepare_probe_data(
        precomputed_acts, class_idx, perform_scr
    )

    assert inputs_train_BLD.shape[0] == len(labels_train_B)

    device = inputs_train_BLD.device
    dtype = inputs_train_BLD.dtype

    running_sum_pos_F = torch.zeros(
        sae.W_dec.data.shape[0], dtype=torch.float32, device=device
    )
    running_sum_neg_F = torch.zeros(
        sae.W_dec.data.shape[0], dtype=torch.float32, device=device
    )
    count_pos = 0
    count_neg = 0

    for i in range(0, inputs_train_BLD.shape[0], sae_batch_size):
        activation_batch_BLD = inputs_train_BLD[i : i + sae_batch_size]
        labels_batch_B = labels_train_B[i : i + sae_batch_size]

        activations_BL = einops.reduce(activation_batch_BLD, "B L D -> B L", "sum")
        nonzero_acts_BL = (activations_BL != 0.0).to(dtype=dtype)
        nonzero_acts_B = einops.reduce(nonzero_acts_BL, "B L -> B", "sum").to(
            torch.float32
        )

        f_BLF = sae.encode(activation_batch_BLD)
        f_BLF = f_BLF * nonzero_acts_BL[:, :, None]  # zero out masked tokens

        # Get the average activation per input. We divide by the number of nonzero activations for the attention mask
        average_sae_acts_BF = (
            einops.reduce(f_BLF, "B L F -> B F", "sum").to(torch.float32)
            / nonzero_acts_B[:, None]
        )

        # Separate positive and negative samples
        pos_mask = labels_batch_B == dataset_info.POSITIVE_CLASS_LABEL
        neg_mask = labels_batch_B == dataset_info.NEGATIVE_CLASS_LABEL

        # Accumulate sums in fp32
        running_sum_pos_F += einops.reduce(
            average_sae_acts_BF[pos_mask], "B F -> F", "sum"
        )
        running_sum_neg_F += einops.reduce(
            average_sae_acts_BF[neg_mask], "B F -> F", "sum"
        )

        count_pos += pos_mask.sum().item()
        count_neg += neg_mask.sum().item()

    # Calculate means in fp32
    average_pos_sae_acts_F = (
        running_sum_pos_F / count_pos if count_pos > 0 else running_sum_pos_F
    )
    average_neg_sae_acts_F = (
        running_sum_neg_F / count_neg if count_neg > 0 else running_sum_neg_F
    )

    # The decoder matrix can be very large, so we move it to the same device as the activations
    average_acts_F = (average_pos_sae_acts_F - average_neg_sae_acts_F).to(dtype)

    probe_weight_D = probe.net.weight.to(dtype=dtype, device=device)
    decoder_weight_DF = sae.W_dec.data.T.to(dtype=dtype, device=device)

    dot_prod_F = (probe_weight_D @ decoder_weight_DF).squeeze()

    if not perform_scr:
        # Only consider activations from the positive class
        average_acts_F.clamp_(min=0.0)

    effects_F = (average_acts_F * dot_prod_F).to(dtype=torch.float32)

    if perform_scr:
        effects_F = effects_F.abs()

    return effects_F


def get_all_node_effects_for_one_sae(
    sae: SAE,
    probes: dict[str, probe_training.Probe],
    chosen_class_indices: list[str],
    perform_scr: bool,
    indirect_effect_acts: dict[str, torch.Tensor],
    sae_batch_size: int,
) -> dict[str, torch.Tensor]:
    node_effects = {}
    for ablated_class_idx in chosen_class_indices:
        node_effects[ablated_class_idx] = get_effects_per_class_precomputed_acts(
            sae,
            probes[ablated_class_idx],
            ablated_class_idx,
            indirect_effect_acts,
            perform_scr,
            sae_batch_size,
        )

    return node_effects


def select_top_n_features(
    effects: torch.Tensor, n: int, class_name: str
) -> torch.Tensor:
    assert n <= effects.numel(), (
        f"n ({n}) must not be larger than the number of features ({effects.numel()}) for ablation class {class_name}"
    )

    # Find non-zero effects
    non_zero_mask = effects != 0
    non_zero_effects = effects[non_zero_mask]
    num_non_zero = non_zero_effects.numel()

    if num_non_zero < n:
        print(
            f"WARNING: only {num_non_zero} non-zero effects found for ablation class {class_name}, which is less than the requested {n}."
        )

    # Select top n or all non-zero effects, whichever is smaller
    k = min(n, num_non_zero)

    if k == 0:
        print(
            f"WARNING: No non-zero effects found for ablation class {class_name}. Returning an empty mask."
        )
        top_n_features = torch.zeros_like(effects, dtype=torch.bool)
    else:
        # Get the indices of the top N effects
        _, top_indices = torch.topk(effects, k)

        # Create a boolean mask tensor
        mask = torch.zeros_like(effects, dtype=torch.bool)
        mask[top_indices] = True

        top_n_features = mask

    return top_n_features


def ablated_precomputed_activations(
    ablation_acts_BLD: torch.Tensor,
    sae: SAE,
    to_ablate: torch.Tensor,
    sae_batch_size: int,
) -> torch.Tensor:
    """NOTE: We don't pass in the attention mask. Thus, we must have already zeroed out all masked tokens in ablation_acts_BLD."""

    all_acts_list_BD = []

    for i in range(0, ablation_acts_BLD.shape[0], sae_batch_size):
        activation_batch_BLD = ablation_acts_BLD[i : i + sae_batch_size]
        dtype = activation_batch_BLD.dtype

        activations_BL = einops.reduce(activation_batch_BLD, "B L D -> B L", "sum")
        nonzero_acts_BL = (activations_BL != 0.0).to(dtype=dtype)
        nonzero_acts_B = einops.reduce(nonzero_acts_BL, "B L -> B", "sum")

        f_BLF = sae.encode(activation_batch_BLD)
        x_hat_BLD = sae.decode(f_BLF)

        error_BLD = activation_batch_BLD - x_hat_BLD

        f_BLF[..., to_ablate] = 0.0  # zero ablation

        modified_acts_BLD = sae.decode(f_BLF) + error_BLD

        # Get the average activation per input. We divide by the number of nonzero activations for the attention mask
        probe_acts_BD = (
            einops.reduce(modified_acts_BLD, "B L D -> B D", "sum")
            / nonzero_acts_B[:, None]
        )
        all_acts_list_BD.append(probe_acts_BD)

    all_acts_BD = torch.cat(all_acts_list_BD, dim=0)

    return all_acts_BD


def get_probe_test_accuracy(
    probes: dict[str, probe_training.Probe],
    all_class_list: list[str],
    all_activations: dict[str, torch.Tensor],
    probe_batch_size: int,
    perform_scr: bool,
) -> dict[str, float]:
    test_accuracies = {}
    for class_name in all_class_list:
        test_acts, test_labels = probe_training.prepare_probe_data(
            all_activations, class_name, perform_scr=perform_scr
        )

        test_acc_probe = probe_training.test_probe_gpu(
            test_acts,
            test_labels,
            probe_batch_size,
            probes[class_name],
        )
        test_accuracies[class_name] = test_acc_probe

    if perform_scr:
        scr_probe_accuracies = get_scr_probe_test_accuracy(
            probes, all_class_list, all_activations, probe_batch_size
        )
        test_accuracies.update(scr_probe_accuracies)

    return test_accuracies


def get_scr_probe_test_accuracy(
    probes: dict[str, probe_training.Probe],
    all_class_list: list[str],
    all_activations: dict[str, torch.Tensor],
    probe_batch_size: int,
) -> dict[str, float]:
    """Tests e.g. male_professor / female_nurse probe on professor / nurse labels"""
    test_accuracies = {}
    for class_name in all_class_list:
        if class_name not in dataset_info.PAIRED_CLASS_KEYS:
            continue
        spurious_class_names = [
            key for key in dataset_info.PAIRED_CLASS_KEYS if key != class_name
        ]
        test_acts, test_labels = probe_training.prepare_probe_data(
            all_activations, class_name, perform_scr=True
        )

        for spurious_class_name in spurious_class_names:
            test_acc_probe = probe_training.test_probe_gpu(
                test_acts,
                test_labels,
                probe_batch_size,
                probes[spurious_class_name],
            )
            combined_class_name = f"{spurious_class_name} probe on {class_name} data"
            test_accuracies[combined_class_name] = test_acc_probe

    return test_accuracies


def perform_feature_ablations(
    probes: dict[str, probe_training.Probe],
    sae: SAE,
    sae_batch_size: int,
    all_test_acts_BLD: dict[str, torch.Tensor],
    node_effects: dict[str, torch.Tensor],
    top_n_values: list[int],
    chosen_classes: list[str],
    probe_batch_size: int,
    perform_scr: bool,
) -> dict[str, dict[int, dict[str, float]]]:
    ablated_class_accuracies = {}
    for ablated_class_name in chosen_classes:
        ablated_class_accuracies[ablated_class_name] = {}
        for top_n in top_n_values:
            selected_features_F = select_top_n_features(
                node_effects[ablated_class_name], top_n, ablated_class_name
            )
            test_acts_ablated = {}
            for evaluated_class_name in all_test_acts_BLD.keys():
                test_acts_ablated[evaluated_class_name] = (
                    ablated_precomputed_activations(
                        all_test_acts_BLD[evaluated_class_name],
                        sae,
                        selected_features_F,
                        sae_batch_size,
                    )
                )

            ablated_class_accuracies[ablated_class_name][top_n] = (
                get_probe_test_accuracy(
                    probes,
                    chosen_classes,
                    test_acts_ablated,
                    probe_batch_size,
                    perform_scr,
                )
            )
    return ablated_class_accuracies


def get_scr_plotting_dict(
    class_accuracies: dict[str, dict[int, dict[str, float]]],
    llm_clean_accs: dict[str, float],
) -> dict[str, float]:
    """raw_results: dict[class_name][threshold][class_name] = float
    llm_clean_accs: dict[class_name] = float
    Returns: dict[metric_name] = float"""

    results = {}
    eval_probe_class_id = "male_professor / female_nurse"

    dirs = [1, 2]

    dir1_class_name = f"{eval_probe_class_id} probe on professor / nurse data"
    dir2_class_name = f"{eval_probe_class_id} probe on male / female data"

    dir1_acc = llm_clean_accs[dir1_class_name]
    dir2_acc = llm_clean_accs[dir2_class_name]

    for dir in dirs:
        if dir == 1:
            ablated_probe_class_id = "male / female"
            eval_data_class_id = "professor / nurse"
        elif dir == 2:
            ablated_probe_class_id = "professor / nurse"
            eval_data_class_id = "male / female"
        else:
            raise ValueError("Invalid dir.")

        for threshold in class_accuracies[ablated_probe_class_id]:
            clean_acc = llm_clean_accs[eval_data_class_id]

            combined_class_name = (
                f"{eval_probe_class_id} probe on {eval_data_class_id} data"
            )

            original_acc = llm_clean_accs[combined_class_name]

            changed_acc = class_accuracies[ablated_probe_class_id][threshold][
                combined_class_name
            ]

            if (clean_acc - original_acc) < 0.001:
                scr_score = 0
            else:
                scr_score = (changed_acc - original_acc) / (clean_acc - original_acc)

            print(
                f"dir: {dir}, original_acc: {original_acc}, clean_acc: {clean_acc}, changed_acc: {changed_acc}, scr_score: {scr_score}"
            )

            metric_key = f"scr_dir{dir}_threshold_{threshold}"

            results[metric_key] = scr_score

            scr_metric_key = f"scr_metric_threshold_{threshold}"
            if dir1_acc < dir2_acc and dir == 1:
                results[scr_metric_key] = scr_score
            elif dir1_acc > dir2_acc and dir == 2:
                results[scr_metric_key] = scr_score

    return results


def create_tpp_plotting_dict(
    class_accuracies: dict[str, dict[int, dict[str, float]]],
    llm_clean_accs: dict[str, float],
) -> tuple[dict[str, float], dict[str, dict[str, list[float]]]]:
    """Calculates TPP metrics for each class and overall averages.

    Args:
        class_accuracies: Nested dict mapping class_name -> threshold -> other_class -> accuracy
        llm_clean_accs: Dict mapping class_name -> clean accuracy

    Returns:
        Tuple containing:
        - Dict mapping metric_name -> value for overall averages
        - Dict mapping class_name -> metric_name -> value
    """
    per_class_results = {}
    overall_results = {}
    classes = list(llm_clean_accs.keys())

    for class_name in classes:
        if " probe on " in class_name:
            raise ValueError("This is SCR, shouldn't be here.")

        class_metrics = {}
        intended_clean_acc = llm_clean_accs[class_name]

        # Calculate metrics for each threshold
        for threshold in class_accuracies[class_name]:
            # Intended differences
            intended_patched_acc = class_accuracies[class_name][threshold][class_name]
            intended_diff = intended_clean_acc - intended_patched_acc

            # Unintended differences for this threshold
            unintended_diffs = []
            for unintended_class in classes:
                if unintended_class == class_name:
                    continue

                unintended_clean_acc = llm_clean_accs[unintended_class]
                unintended_patched_acc = class_accuracies[class_name][threshold][
                    unintended_class
                ]
                unintended_diff = unintended_clean_acc - unintended_patched_acc
                unintended_diffs.append(unintended_diff)

            avg_unintended = sum(unintended_diffs) / len(unintended_diffs)
            avg_diff = intended_diff - avg_unintended

            # Store with original key format
            class_metrics[f"tpp_threshold_{threshold}_total_metric"] = avg_diff
            class_metrics[f"tpp_threshold_{threshold}_intended_diff_only"] = (
                intended_diff
            )
            class_metrics[f"tpp_threshold_{threshold}_unintended_diff_only"] = (
                avg_unintended
            )

        per_class_results[class_name] = class_metrics

    # Calculate overall averages across classes
    # First, determine all metric keys from the first class
    metric_keys = next(iter(per_class_results.values())).keys()

    for metric_key in metric_keys:
        values = [
            class_metrics[metric_key] for class_metrics in per_class_results.values()
        ]
        overall_results[metric_key] = sum(values) / len(values)

    return overall_results, per_class_results


def get_dataset_activations(
    dataset_name: str,
    config: ScrAndTppEvalConfig,
    model: HookedTransformer,
    llm_batch_size: int,
    layer: int,
    hook_point: str,
    device: str,
    chosen_classes: list[str],
    column1_vals: tuple[str, str] | None = None,
    column2_vals: tuple[str, str] | None = None,
) -> tuple[dict[str, torch.Tensor], dict[str, torch.Tensor]]:
    train_data, test_data = dataset_creation.get_train_test_data(
        dataset_name,
        config.perform_scr,
        config.train_set_size,
        config.test_set_size,
        config.random_seed,
        column1_vals,
        column2_vals,
    )

    if not config.perform_scr:
        train_data = dataset_utils.filter_dataset(train_data, chosen_classes)
        test_data = dataset_utils.filter_dataset(test_data, chosen_classes)

    train_data = dataset_utils.tokenize_data_dictionary(
        train_data,
        model.tokenizer,  # type: ignore
        config.context_length,
        device,
    )
    test_data = dataset_utils.tokenize_data_dictionary(
        test_data,
        model.tokenizer,  # type: ignore
        config.context_length,
        device,
    )

    all_train_acts_BLD = activation_collection.get_all_llm_activations(
        train_data,
        model,
        llm_batch_size,
        layer,
        hook_point,
        mask_bos_pad_eos_tokens=True,
    )
    all_test_acts_BLD = activation_collection.get_all_llm_activations(
        test_data,
        model,
        llm_batch_size,
        layer,
        hook_point,
        mask_bos_pad_eos_tokens=True,
    )

    return all_train_acts_BLD, all_test_acts_BLD


def run_eval_single_dataset(
    dataset_name: str,
    config: ScrAndTppEvalConfig,
    sae: SAE,
    model: HookedTransformer,
    layer: int,
    hook_point: str,
    device: str,
    artifacts_folder: str,
    save_activations: bool = True,
    column1_vals: tuple[str, str] | None = None,
) -> tuple[dict[str, dict[str, dict[int, dict[str, float]]]], dict[str, float]]:
    """Return dict is of the form:
    dict[ablated_class_name][threshold][measured_acc_class_name] = float

    config: eval_config.EvalConfig contains all hyperparameters to reproduce the evaluation.
    It is saved in the results_dict for reproducibility."""

    column2_vals = COLUMN2_VALS_LOOKUP[dataset_name]

    if not config.perform_scr:
        chosen_classes = dataset_info.chosen_classes_per_dataset[dataset_name]
        activations_filename = f"{dataset_name}_activations.pt".replace("/", "_")
        probes_filename = f"{dataset_name}_probes.pkl".replace("/", "_")
    else:
        chosen_classes = list(dataset_info.PAIRED_CLASS_KEYS.keys())
        activations_filename = f"{dataset_name}_{column1_vals[0]}_{column1_vals[1]}_activations.pt".replace(  # type: ignore
            "/", "_"
        )
        probes_filename = (
            f"{dataset_name}_{column1_vals[0]}_{column1_vals[1]}_probes.pkl".replace(  # type: ignore
                "/", "_"
            )
        )

    activations_path = os.path.join(artifacts_folder, activations_filename)
    probes_path = os.path.join(artifacts_folder, probes_filename)

    if not os.path.exists(activations_path):
        if config.lower_vram_usage:
            model = model.to(device)  # type: ignore
        all_train_acts_BLD, all_test_acts_BLD = get_dataset_activations(
            dataset_name,
            config,
            model,
            config.llm_batch_size,
            layer,
            hook_point,
            device,
            chosen_classes,
            column1_vals,
            column2_vals,
        )
        if config.lower_vram_usage:
            model = model.to("cpu")  # type: ignore

        all_meaned_train_acts_BD = (
            activation_collection.create_meaned_model_activations(all_train_acts_BLD)
        )
        all_meaned_test_acts_BD = activation_collection.create_meaned_model_activations(
            all_test_acts_BLD
        )

        torch.set_grad_enabled(True)

        llm_probes, llm_test_accuracies = probe_training.train_probe_on_activations(
            all_meaned_train_acts_BD,
            all_meaned_test_acts_BD,
            select_top_k=None,
            use_sklearn=False,
            batch_size=config.probe_train_batch_size,
            epochs=config.probe_epochs,
            lr=config.probe_lr,
            perform_scr=config.perform_scr,
            early_stopping_patience=config.early_stopping_patience,
            l1_penalty=config.probe_l1_penalty,
        )

        torch.set_grad_enabled(False)

        llm_test_accuracies = get_probe_test_accuracy(
            llm_probes,  # type: ignore
            chosen_classes,
            all_meaned_test_acts_BD,
            config.probe_test_batch_size,
            config.perform_scr,
        )

        acts = {
            "train": all_train_acts_BLD,
            "test": all_test_acts_BLD,
        }

        llm_probes_dict = {
            "llm_probes": llm_probes,
            "llm_test_accuracies": llm_test_accuracies,
        }

        if save_activations:
            torch.save(acts, activations_path)
            with open(probes_path, "wb") as f:
                pickle.dump(llm_probes_dict, f)
    else:
        if config.lower_vram_usage:
            model = model.to("cpu")  # type: ignore
        print(f"Loading activations from {activations_path}")
        acts = torch.load(activations_path)
        all_train_acts_BLD = acts["train"]
        all_test_acts_BLD = acts["test"]

        print(f"Loading probes from {probes_path}")
        with open(probes_path, "rb") as f:
            llm_probes_dict = pickle.load(f)

        llm_probes = llm_probes_dict["llm_probes"]
        llm_test_accuracies = llm_probes_dict["llm_test_accuracies"]

    torch.set_grad_enabled(False)

    sae_node_effects = get_all_node_effects_for_one_sae(
        sae,
        llm_probes,  # type: ignore
        chosen_classes,
        config.perform_scr,
        all_train_acts_BLD,
        config.sae_batch_size,
    )

    ablated_class_accuracies = perform_feature_ablations(
        llm_probes,  # type: ignore
        sae,
        config.sae_batch_size,
        all_test_acts_BLD,
        sae_node_effects,
        config.n_values,
        chosen_classes,
        config.probe_test_batch_size,
        config.perform_scr,
    )

    return ablated_class_accuracies, llm_test_accuracies  # type: ignore


def run_eval_single_sae(
    config: ScrAndTppEvalConfig,
    sae: SAE,
    model: HookedTransformer,
    device: str,
    artifacts_folder: str,
    save_activations: bool = True,
) -> tuple[dict[str, float | dict[str, float]], dict]:
    """hook_point: str is transformer lens format. example: f'blocks.{layer}.hook_resid_post'
    By default, we save activations for all datasets, and then reuse them for each sae.
    This is important to avoid recomputing activations for each SAE, and to ensure that the same activations are used for all SAEs.
    However, it can use 10s of GBs of disk space."""

    random.seed(config.random_seed)
    torch.manual_seed(config.random_seed)

    os.makedirs(artifacts_folder, exist_ok=True)

    dataset_results = {}
    per_dataset_class_results = {}

    averaging_names = []

    for dataset_name in config.dataset_names:
        if config.perform_scr:
            column1_vals_list = config.column1_vals_lookup[dataset_name]
            for column1_vals in column1_vals_list:
                run_name = f"{dataset_name}_scr_{column1_vals[0]}_{column1_vals[1]}"
                raw_results, llm_clean_accs = run_eval_single_dataset(
                    dataset_name,
                    config,
                    sae,
                    model,
                    sae.cfg.hook_layer,
                    sae.cfg.hook_name,
                    device,
                    artifacts_folder,
                    save_activations,
                    column1_vals,
                )

                processed_results = get_scr_plotting_dict(raw_results, llm_clean_accs)  # type: ignore

                dataset_results[f"{run_name}_results"] = processed_results

                averaging_names.append(run_name)

        else:
            run_name = f"{dataset_name}_tpp"
            raw_results, llm_clean_accs = run_eval_single_dataset(
                dataset_name,
                config,
                sae,
                model,
                sae.cfg.hook_layer,
                sae.cfg.hook_name,
                device,
                artifacts_folder,
                save_activations,
            )

            processed_results, per_class_results = create_tpp_plotting_dict(
                raw_results,  # type: ignore
                llm_clean_accs,
            )
            dataset_results[f"{run_name}_results"] = processed_results
            per_dataset_class_results[dataset_name] = per_class_results

            averaging_names.append(run_name)

    results_dict = general_utils.average_results_dictionaries(
        dataset_results, averaging_names
    )
    results_dict.update(dataset_results)

    if config.lower_vram_usage:
        model = model.to(device)  # type: ignore

    return results_dict, per_dataset_class_results  # type: ignore


def run_eval(
    config: ScrAndTppEvalConfig,
    selected_saes: list[tuple[str, SAE]] | list[tuple[str, str]],
    device: str,
    output_path: str,
    force_rerun: bool = False,
    clean_up_activations: bool = False,
    save_activations: bool = True,
    artifacts_path: str = "artifacts",
):
    """
    selected_saes is a list of either tuples of (sae_lens release, sae_lens id) or (sae_name, SAE object)

    If clean_up_activations is True, which means that the activations are deleted after the evaluation is done.
    You may want to use this because activations for all datasets can easily be 10s of GBs.
    Return dict is a dict of SAE name: evaluation results for that SAE."""
    eval_instance_id = get_eval_uuid()
    sae_lens_version = get_sae_lens_version()
    sae_bench_commit_hash = get_sae_bench_version()

    if config.perform_scr:
        eval_type = EVAL_TYPE_ID_SCR
    else:
        eval_type = EVAL_TYPE_ID_TPP
    output_path = os.path.join(output_path, eval_type)
    os.makedirs(output_path, exist_ok=True)

    artifacts_folder = None

    results_dict = {}

    llm_dtype = general_utils.str_to_dtype(config.llm_dtype)

    model = HookedTransformer.from_pretrained_no_processing(
        config.model_name, device=device, dtype=llm_dtype
    )

    for sae_release, sae_object_or_id in tqdm(
        selected_saes, desc="Running SAE evaluation on all selected SAEs"
    ):
        sae_id, sae, sparsity = general_utils.load_and_format_sae(
            sae_release, sae_object_or_id, device
        )  # type: ignore
        sae = sae.to(device=device, dtype=llm_dtype)

        sae_result_path = general_utils.get_results_filepath(
            output_path, sae_release, sae_id
        )

        if os.path.exists(sae_result_path) and not force_rerun:
            print(f"Skipping {sae_release}_{sae_id} as results already exist")
            continue

        artifacts_folder = os.path.join(
            artifacts_path,
            eval_type,
            config.model_name,
            sae.cfg.hook_name,
        )

        scr_or_tpp_results, per_dataset_class_results = run_eval_single_sae(
            config,
            sae,
            model,
            device,
            artifacts_folder,
            save_activations,
        )
        if eval_type == EVAL_TYPE_ID_SCR:
            eval_output = ScrEvalOutput(
                eval_type_id=eval_type,
                eval_config=config,
                eval_id=eval_instance_id,
                datetime_epoch_millis=int(datetime.now().timestamp() * 1000),
                eval_result_metrics=ScrMetricCategories(
                    scr_metrics=ScrMetrics(
                        **{
                            k: v
                            for k, v in scr_or_tpp_results.items()
                            if not isinstance(v, dict)
                        }
                    )
                ),
                eval_result_details=[
                    ScrResultDetail(
                        dataset_name=dataset_name,
                        **result,
                    )
                    for dataset_name, result in scr_or_tpp_results.items()
                    if isinstance(result, dict)
                ],
                sae_bench_commit_hash=sae_bench_commit_hash,
                sae_lens_id=sae_id,
                sae_lens_release_id=sae_release,
                sae_lens_version=sae_lens_version,
                sae_cfg_dict=asdict(sae.cfg),
            )
        elif eval_type == EVAL_TYPE_ID_TPP:
            eval_output = TppEvalOutput(
                eval_type_id=eval_type,
                eval_config=config,
                eval_id=eval_instance_id,
                datetime_epoch_millis=int(datetime.now().timestamp() * 1000),
                eval_result_metrics=TppMetricCategories(
                    tpp_metrics=TppMetrics(
                        **{
                            k: v
                            for k, v in scr_or_tpp_results.items()
                            if not isinstance(v, dict)
                        }
                    )
                ),
                eval_result_details=[
                    TppResultDetail(
                        dataset_name=dataset_name,
                        **result,
                    )
                    for dataset_name, result in scr_or_tpp_results.items()
                    if isinstance(result, dict)
                ],
                eval_result_unstructured=per_dataset_class_results,
                sae_bench_commit_hash=sae_bench_commit_hash,
                sae_lens_id=sae_id,
                sae_lens_release_id=sae_release,
                sae_lens_version=sae_lens_version,
                sae_cfg_dict=asdict(sae.cfg),
            )
        else:
            raise ValueError(f"Invalid eval type: {eval_type}")

        results_dict[f"{sae_release}_{sae_id}"] = asdict(eval_output)

        eval_output.to_json_file(sae_result_path, indent=2)

        gc.collect()
        torch.cuda.empty_cache()

    if clean_up_activations:
        if artifacts_folder is not None and os.path.exists(artifacts_folder):
            shutil.rmtree(artifacts_folder)

    return results_dict


def create_config_and_selected_saes(
    args,
) -> tuple[ScrAndTppEvalConfig, list[tuple[str, str]]]:
    config = ScrAndTppEvalConfig(
        model_name=args.model_name,
        perform_scr=args.perform_scr,
    )

    if args.llm_batch_size is not None:
        config.llm_batch_size = args.llm_batch_size
    else:
        config.llm_batch_size = activation_collection.LLM_NAME_TO_BATCH_SIZE[
            config.model_name
        ]

    if args.llm_dtype is not None:
        config.llm_dtype = args.llm_dtype
    else:
        config.llm_dtype = activation_collection.LLM_NAME_TO_DTYPE[config.model_name]

    if args.random_seed is not None:
        config.random_seed = args.random_seed

    if args.lower_vram_usage:
        config.lower_vram_usage = True

    if args.sae_batch_size is not None:
        config.sae_batch_size = args.sae_batch_size

    selected_saes = get_saes_from_regex(args.sae_regex_pattern, args.sae_block_pattern)
    assert len(selected_saes) > 0, "No SAEs selected"

    releases = set([release for release, _ in selected_saes])

    print(f"Selected SAEs from releases: {releases}")

    for release, sae in selected_saes:
        print(f"Sample SAEs: {release}, {sae}")

    return config, selected_saes


def arg_parser():
    parser = argparse.ArgumentParser(description="Run SCR or TPP evaluation")
    parser.add_argument("--random_seed", type=int, default=None, help="Random seed")
    parser.add_argument("--model_name", type=str, required=True, help="Model name")
    parser.add_argument(
        "--sae_regex_pattern",
        type=str,
        required=True,
        help="Regex pattern for SAE selection",
    )
    parser.add_argument(
        "--sae_block_pattern",
        type=str,
        required=True,
        help="Regex pattern for SAE block selection",
    )
    parser.add_argument(
        "--output_folder",
        type=str,
        default="eval_results",
        help="Output folder",
    )
    parser.add_argument(
        "--artifacts_path",
        type=str,
        default="artifacts",
        help="Path to save artifacts",
    )
    parser.add_argument(
        "--force_rerun", action="store_true", help="Force rerun of experiments"
    )
    parser.add_argument(
        "--clean_up_activations",
        action="store_true",
        help="Clean up activations after evaluation",
    )
    parser.add_argument(
        "--save_activations",
        action="store_false",
        help="Save the generated LLM activations for later use",
    )

    def str_to_bool(value):
        if value.lower() in ("true", "false"):
            return value.lower() == "true"
        raise argparse.ArgumentTypeError("Boolean value expected.")

    parser.add_argument(
        "--perform_scr",
        type=str_to_bool,
        required=True,
        help="If true, do Spurious Correlation Removal (SCR). If false, do TPP.",
    )
    parser.add_argument(
        "--llm_batch_size",
        type=int,
        default=None,
        help="Batch size for LLM. If None, will be populated using LLM_NAME_TO_BATCH_SIZE",
    )
    parser.add_argument(
        "--llm_dtype",
        type=str,
        default=None,
        choices=[None, "float32", "float64", "float16", "bfloat16"],
        help="Data type for LLM. If None, will be populated using LLM_NAME_TO_DTYPE",
    )
    parser.add_argument(
        "--sae_batch_size",
        type=int,
        default=None,
        help="Batch size for SAE. If None, will be populated using default config value",
    )
    parser.add_argument(
        "--lower_vram_usage",
        action="store_true",
        help="Lower GPU memory usage by moving model to CPU when not required. Useful on 1M width SAEs. Will be slower and require more system memory.",
    )

    return parser


if __name__ == "__main__":
    """
    Example pythia-70m usage:
    python evals/scr_and_tpp/main.py \
    --sae_regex_pattern "sae_bench_pythia70m_sweep_standard_ctx128_0712" \
    --sae_block_pattern "blocks.4.hook_resid_post__trainer_10" \
    --model_name pythia-70m-deduped \
    --perform_scr true

    Example Gemma-2-2B SAE Bench usage:
    python evals/scr_and_tpp/main.py \
    --sae_regex_pattern "sae_bench_gemma-2-2b_topk_width-2pow14_date-1109" \
    --sae_block_pattern "blocks.19.hook_resid_post__trainer_2" \
    --model_name gemma-2-2b \
    --perform_scr true

    Example Gemma-2-2B Gemma-Scope usage:
    python evals/scr_and_tpp/main.py \
    --sae_regex_pattern "gemma-scope-2b-pt-res" \
    --sae_block_pattern "layer_20/width_16k/average_l0_139" \
    --model_name gemma-2-2b \
    --perform_scr true
    """
    args = arg_parser().parse_args()
    device = general_utils.setup_environment()

    start_time = time.time()

    config, selected_saes = create_config_and_selected_saes(args)

    print(selected_saes)

    # create output folder
    os.makedirs(args.output_folder, exist_ok=True)

    # run the evaluation on all selected SAEs
    results_dict = run_eval(
        config,
        selected_saes,
        device,
        args.output_folder,
        args.force_rerun,
        args.clean_up_activations,
        args.save_activations,
        artifacts_path=args.artifacts_path,
    )

    end_time = time.time()

    print(f"Finished evaluation in {end_time - start_time} seconds")


# Use this code snippet to use custom SAE objects
# if __name__ == "__main__":
#     import sae_bench.custom_saes.identity_sae as identity_sae
#     import sae_bench.custom_saes.jumprelu_sae as jumprelu_sae

#     """
#     python evals/scr_and_tpp/main.py
#     """
#     device = general_utils.setup_environment()

#     start_time = time.time()

#     random_seed = 42
#     output_folder = "eval_results"
#     perform_scr = True

#     model_name = "gemma-2-2b"
#     hook_layer = 20

#     repo_id = "google/gemma-scope-2b-pt-res"
#     filename = f"layer_{hook_layer}/width_16k/average_l0_71/params.npz"
#     sae = jumprelu_sae.load_jumprelu_sae(repo_id, filename, hook_layer)
#     selected_saes = [(f"{repo_id}_{filename}_gemmascope_sae", sae)]

#     config = ScrAndTppEvalConfig(
#         random_seed=random_seed,
#         model_name=model_name,
#         perform_scr=perform_scr,
#     )

#     config.llm_batch_size = activation_collection.LLM_NAME_TO_BATCH_SIZE[config.model_name]
#     config.llm_dtype = activation_collection.LLM_NAME_TO_DTYPE[config.model_name]

#     # create output folder
#     os.makedirs(output_folder, exist_ok=True)

#     # run the evaluation on all selected SAEs
#     results_dict = run_eval(
#         config,
#         selected_saes,
#         device,
#         output_folder,
#         force_rerun=True,
#         clean_up_activations=False,
#         save_activations=True,
#     )

#     end_time = time.time()

#     print(f"Finished evaluation in {end_time - start_time} seconds")



import argparse
import gc
import os
import random
import time
from dataclasses import asdict
from datetime import datetime

import torch
from sae_lens import SAE
from tqdm import tqdm
from transformers import (
    AutoModelForCausalLM,
    AutoTokenizer,
    BatchEncoding,
    PreTrainedModel,
)

import sae_bench.evals.ravel.intervention as intervention
import sae_bench.evals.ravel.mdbm as mdbm
import sae_bench.sae_bench_utils.activation_collection as activation_collection
import sae_bench.sae_bench_utils.general_utils as general_utils
from sae_bench.evals.ravel.eval_config import RAVELEvalConfig
from sae_bench.evals.ravel.eval_output import (
    EVAL_TYPE_ID_RAVEL,
    RAVELEvalOutput,
    RAVELMetricCategories,
    RAVELMetricResults,
)
from sae_bench.evals.ravel.generation import custom_left_padding
from sae_bench.evals.ravel.instance import (
    RAVELFilteredDataset,
    RAVELInstance,
    get_instance_name,
)
from sae_bench.evals.ravel.intervention import get_prompt_pairs
from sae_bench.sae_bench_utils import (
    get_eval_uuid,
    get_sae_bench_version,
    get_sae_lens_version,
)
from sae_bench.sae_bench_utils.sae_selection_utils import (
    get_saes_from_regex,
)

# For our initial experiments, we only used transformer_lens and the shorter LLM names
# For RAVEL, we use HF transformers, which requires the full model name
# So, it's the shorter name: full name map
LLM_NAME_MAP = {
    "gemma-2-2b": "google/gemma-2-2b",
    "pythia-160m-deduped": "EleutherAI/pythia-160m-deduped",
}


def create_dataloaders(
    cause_base_prompts,
    cause_source_prompts,
    iso_base_prompts,
    iso_source_prompts,
    model: PreTrainedModel,
    eval_config: RAVELEvalConfig,
    train_test_split: float,
):
    """
    Create train and validation dataloaders from prompt pairs.

    Args:
        cause_base_prompts: List of base prompts
        cause_source_prompts: List of source prompts
        model: The model (used for device information)
        eval_config: Configuration for evaluation
        train_test_split: Ratio of data to use for training (default: 0.5)

    Returns:
        train_loader: Dataloader for training
        val_loader: Dataloader for validation
    """
    # NOTE: Pay very close attention to the order of the arguments here and the difference between cause and iso
    # This determines the labels that are used for cause and iso
    formatted_cause_pairs = []
    for base, source in zip(cause_base_prompts, cause_source_prompts):
        formatted_cause_pairs.append(
            (
                base.input_ids,
                source.input_ids,
                base.attention_mask,
                source.attention_mask,
                base.final_entity_token_pos,
                source.final_entity_token_pos,
                base.first_generated_token_id,
                source.first_generated_token_id,
                base.text,
                source.attribute_label,  # NOTE: We want to change the label to source for cause
            )
        )

    formatted_iso_pairs = []
    for base, source in zip(iso_base_prompts, iso_source_prompts):
        formatted_iso_pairs.append(
            (
                base.input_ids,
                source.input_ids,
                base.attention_mask,
                source.attention_mask,
                base.final_entity_token_pos,
                source.final_entity_token_pos,
                base.first_generated_token_id,
                base.first_generated_token_id,  # NOTE: We want the label to remain as base for iso
                base.text,
                base.attribute_label,
            )
        )

    all_formatted_pairs = formatted_cause_pairs + formatted_iso_pairs
    random.shuffle(all_formatted_pairs)

    # Split into train and validation sets
    total_pairs = len(all_formatted_pairs)
    train_size = int(total_pairs * train_test_split)

    train_pairs = all_formatted_pairs[:train_size]
    val_pairs = all_formatted_pairs[train_size:]

    print(
        f"Created {len(train_pairs)} training pairs and {len(val_pairs)} validation pairs"
    )

    # Create dataloaders
    train_loader = create_dataloader_from_pairs(train_pairs, model, eval_config)
    val_loader = create_dataloader_from_pairs(val_pairs, model, eval_config)

    return train_loader, val_loader


def create_dataloader_from_pairs(formatted_pairs, model, eval_config):
    """
    Create a dataloader from formatted prompt pairs.

    Args:
        formatted_pairs: List of formatted prompt pairs
        model: The model (used for device information)
        eval_config: Configuration for evaluation

    Returns:
        dataloader: List of batched data
    """
    dataloader = []
    num_batches = len(formatted_pairs) // eval_config.llm_batch_size

    tokenizer = AutoTokenizer.from_pretrained(eval_config.model_name)

    max_base_len = max(len(pair[0]) for pair in formatted_pairs)
    max_source_len = max(len(pair[1]) for pair in formatted_pairs)

    for batch_idx in range(num_batches):
        batch_start = batch_idx * eval_config.llm_batch_size
        batch_end = batch_start + eval_config.llm_batch_size
        batch_data = formatted_pairs[batch_start:batch_end]

        base_tokens_BL = []
        source_tokens_BL = []
        base_attn_mask_BL = []
        source_attn_mask_BL = []
        base_pos_B = []
        source_pos_B = []
        base_pred_B = []
        source_pred_B = []
        base_text_str = []
        base_label_str = []

        for (
            base_tokens_L,
            source_tokens_L,
            base_attn_mask_L,
            source_attn_mask_L,
            base_pos,
            source_pos,
            base_pred,
            source_pred,
            base_text,
            base_label,
        ) in batch_data:
            base_tokens_BL.append(base_tokens_L)
            source_tokens_BL.append(source_tokens_L)
            base_attn_mask_BL.append(base_attn_mask_L)
            source_attn_mask_BL.append(source_attn_mask_L)
            base_pos_B.append(base_pos)
            source_pos_B.append(source_pos)
            base_pred_B.append(base_pred)
            source_pred_B.append(source_pred)
            base_text_str.append(base_text)
            base_label_str.append(base_label)

        base_tokens_BL, base_attn_mask_BL = custom_left_padding(
            tokenizer, base_tokens_BL, max_base_len
        )
        source_tokens_BL, source_attn_mask_BL = custom_left_padding(
            tokenizer, source_tokens_BL, max_source_len
        )

        base_tokens_BL = base_tokens_BL.to(model.device)
        base_attn_mask_BL = base_attn_mask_BL.to(model.device)
        source_tokens_BL = source_tokens_BL.to(model.device)
        source_attn_mask_BL = source_attn_mask_BL.to(model.device)

        base_pos_B = torch.tensor(base_pos_B).to(model.device)
        source_pos_B = torch.tensor(source_pos_B).to(model.device)
        base_pred_B = torch.tensor(base_pred_B).to(model.device)
        source_pred_B = torch.tensor(source_pred_B).to(model.device)

        base_encoding_BL = BatchEncoding(
            {
                "input_ids": base_tokens_BL,
                "attention_mask": base_attn_mask_BL,
            }
        )
        source_encoding_BL = BatchEncoding(
            {
                "input_ids": source_tokens_BL,
                "attention_mask": source_attn_mask_BL,
            }
        )

        dataloader.append(
            (
                base_encoding_BL,
                source_encoding_BL,
                base_pos_B,
                source_pos_B,
                base_pred_B,
                source_pred_B,
                base_text_str,
                base_label_str,
            )
        )

    print(f"Created dataloader with {len(dataloader)} batches")
    return dataloader


def run_eval_single_cause_attribute(
    dataset,
    cause_attribute: str,
    iso_attributes: list[str],
    config: RAVELEvalConfig,
    sae: SAE,
    model: PreTrainedModel,
) -> dict[str, float]:
    tokenizer = AutoTokenizer.from_pretrained(config.model_name)

    cause_base_prompts, cause_source_prompts = get_prompt_pairs(
        dataset=dataset,
        base_attribute=cause_attribute,
        source_attribute=cause_attribute,
        n_interventions=config.num_pairs_per_attribute,
    )

    iso_base_prompts = []
    iso_source_prompts = []
    for iso_attr in iso_attributes:
        attr_base_prompts, attr_source_prompts = get_prompt_pairs(
            dataset=dataset,
            base_attribute=iso_attr,
            source_attribute=iso_attr,
            n_interventions=config.num_pairs_per_attribute,
        )
        iso_base_prompts.extend(attr_base_prompts)
        iso_source_prompts.extend(attr_source_prompts)

    combined = list(zip(iso_base_prompts, iso_source_prompts))
    random.shuffle(combined)
    iso_base_prompts, iso_source_prompts = zip(*combined)

    # Truncate to match the length of cause prompts
    cause_length = len(cause_base_prompts)
    iso_base_prompts = list(iso_base_prompts[:cause_length])
    iso_source_prompts = list(iso_source_prompts[:cause_length])

    print(
        f"Using {len(cause_base_prompts)} cause prompt pairs and {len(iso_base_prompts)} ISO prompt pairs"
    )

    train_loader, val_loader = create_dataloaders(
        cause_base_prompts,
        cause_source_prompts,
        iso_base_prompts,
        iso_source_prompts,
        model,
        config,
        train_test_split=config.train_test_split,
    )

    torch.cuda.empty_cache()
    gc.collect()

    trained_mdbm = mdbm.train_mdbm(
        model,
        tokenizer,
        config,
        sae,
        train_loader=train_loader,
        val_loader=val_loader,
        verbose=False,
        train_mdas=config.train_mdas,
    )

    torch.cuda.empty_cache()
    gc.collect()

    iso_score, cause_score = intervention.generate_batched_interventions(
        model,
        trained_mdbm,
        tokenizer,
        val_loader,  # type: ignore
        max_new_tokens=config.n_generated_tokens,
    )

    return {
        "cause_score": cause_score,
        "isolation_score": iso_score,
        "disentangle_score": (cause_score + iso_score) / 2,
    }


def run_eval_single_dataset(
    entity_class: str,
    config: RAVELEvalConfig,
    sae: SAE,
    model: PreTrainedModel,
) -> tuple[dict[str, float], dict]:
    """config: eval_config.EvalConfig contains all hyperparameters to reproduce the evaluation.
    It is saved in the results_dict for reproducibility."""

    tokenizer = AutoTokenizer.from_pretrained(config.model_name)

    filtered_dataset_filename = get_instance_name(
        entity_class,
        config.model_name,
        config.full_dataset_downsample,
        config.top_n_entities,
    )
    filtered_dataset_path = os.path.join(config.artifact_dir, filtered_dataset_filename)

    if not os.path.exists(filtered_dataset_path):
        orig_batch_size = config.llm_batch_size
        # Generations use much less memory than training the MDBM
        config.llm_batch_size = orig_batch_size * 8
        full_dataset = RAVELInstance.create_from_files(
            config=config,
            entity_type=entity_class,
            tokenizer=tokenizer,
            data_dir=config.artifact_dir,
            model=model,
            model_name=config.model_name,
            attribute_types=config.entity_attribute_selection[entity_class],
            downsample=config.full_dataset_downsample,
        )
        config.llm_batch_size = orig_batch_size

        # Create filtered dataset.
        full_dataset.create_and_save_filtered_dataset(
            artifact_dir=config.artifact_dir,
            top_n_entities=config.top_n_entities,
        )

    # Test loading the filtered dataset.
    dataset = RAVELFilteredDataset.load(filtered_dataset_path)
    ##########################

    attributes = config.entity_attribute_selection[entity_class]

    results_dict = {"cause_score": [], "isolation_score": [], "disentangle_score": []}
    per_class_results_dict = {}

    for cause_attribute in attributes:
        iso_attributes = [attr for attr in attributes if attr != cause_attribute]

        gc.collect()
        torch.cuda.empty_cache()

        mdbm_results = run_eval_single_cause_attribute(
            dataset,
            cause_attribute,
            iso_attributes,
            config,
            sae,
            model,
        )

        print(mdbm_results)

        results_dict["cause_score"].append(mdbm_results["cause_score"])
        results_dict["isolation_score"].append(mdbm_results["isolation_score"])
        results_dict["disentangle_score"].append(mdbm_results["disentangle_score"])

        per_class_results_dict[f"{entity_class}_{cause_attribute}"] = mdbm_results

    for key in results_dict.keys():
        results_dict[key] = sum(results_dict[key]) / len(results_dict[key])  # type: ignore

    return results_dict, per_class_results_dict  # type: ignore


def run_eval_single_sae(
    config: RAVELEvalConfig,
    sae: SAE,
    model: PreTrainedModel,
    device: str,
    artifacts_folder: str,
) -> tuple[dict[str, float | dict[str, float]], dict]:
    """NOTE: This is currently setup for Transformers, not TransformerLens models."""

    random.seed(config.random_seed)
    torch.manual_seed(config.random_seed)
    os.makedirs(artifacts_folder, exist_ok=True)
    torch.set_grad_enabled(True)

    results_dict = {}

    dataset_results = {}
    per_class_dict = {}
    for entity_class in config.entity_attribute_selection.keys():
        (
            dataset_results[f"{entity_class}_results"],
            per_class_dict[f"{entity_class}_results"],
        ) = run_eval_single_dataset(
            entity_class,
            config,
            sae,
            model,
        )

    results_dict = general_utils.average_results_dictionaries(
        dataset_results, list(config.entity_attribute_selection.keys())
    )

    for entity_class, dataset_result in dataset_results.items():
        results_dict[f"{entity_class}"] = dataset_result

    return results_dict, per_class_dict  # type: ignore


def run_eval(
    config: RAVELEvalConfig,
    selected_saes: list[tuple[str, SAE]] | list[tuple[str, str]],
    device: str,
    output_path: str,
    force_rerun: bool = False,
    artifacts_path: str = "artifacts",
):
    """
    selected_saes is a list of either tuples of (sae_lens release, sae_lens id) or (sae_name, SAE object)

    Return dict is a dict of SAE name: evaluation results for that SAE."""
    eval_instance_id = get_eval_uuid()
    sae_lens_version = get_sae_lens_version()
    sae_bench_commit_hash = get_sae_bench_version()

    gc.collect()
    torch.cuda.empty_cache()

    artifacts_folder = None
    os.makedirs(output_path, exist_ok=True)

    results_dict = {}

    llm_dtype = general_utils.str_to_dtype(config.llm_dtype)
    config.model_name = LLM_NAME_MAP[config.model_name]

    if "gemma" in config.model_name:
        model_kwargs = {"attn_implementation": "eager"}
    else:
        model_kwargs = {}

    model = AutoModelForCausalLM.from_pretrained(
        config.model_name,
        device_map=device,
        torch_dtype=llm_dtype,
        **model_kwargs,
    )

    for sae_release, sae_object_or_id in tqdm(
        selected_saes, desc="Running SAE evaluation on all selected SAEs"
    ):
        sae_id, sae, sparsity = general_utils.load_and_format_sae(
            sae_release, sae_object_or_id, device
        )  # type: ignore
        sae = sae.to(device=device, dtype=llm_dtype)

        if config.train_mdas:
            sae_release = "mdas"
            sae_id = "mdas"
            assert len(selected_saes) == 1

        sae_result_path = general_utils.get_results_filepath(
            output_path, sae_release, sae_id
        )

        if os.path.exists(sae_result_path) and not force_rerun:
            print(f"Skipping {sae_release}_{sae_id} as results already exist")
            continue

        artifacts_folder = os.path.join(
            artifacts_path,
            EVAL_TYPE_ID_RAVEL,
            config.model_name,
            sae.cfg.hook_name,
        )

        eval_results, per_class_dict = run_eval_single_sae(
            config,
            sae,
            model,
            device,
            artifacts_folder,
        )
        eval_output = RAVELEvalOutput(
            eval_config=config,
            eval_id=eval_instance_id,
            datetime_epoch_millis=int(datetime.now().timestamp() * 1000),
            eval_result_metrics=RAVELMetricCategories(
                ravel=RAVELMetricResults(
                    disentanglement_score=eval_results["disentangle_score"],  # type: ignore
                    cause_score=eval_results["cause_score"],  # type: ignore
                    isolation_score=eval_results["isolation_score"],  # type: ignore
                )
            ),
            eval_result_details=[],
            eval_result_unstructured=per_class_dict,
            sae_bench_commit_hash=sae_bench_commit_hash,
            sae_lens_id=sae_id,
            sae_lens_release_id=sae_release,
            sae_lens_version=sae_lens_version,
            sae_cfg_dict=asdict(sae.cfg),
        )

        results_dict[f"{sae_release}_{sae_id}"] = asdict(eval_output)

        eval_output.to_json_file(sae_result_path, indent=2)

        gc.collect()
        torch.cuda.empty_cache()

    return results_dict


def create_config_and_selected_saes(
    args,
) -> tuple[RAVELEvalConfig, list[tuple[str, str]]]:
    config = RAVELEvalConfig(
        model_name=args.model_name,
    )

    if args.llm_batch_size is not None:
        config.llm_batch_size = args.llm_batch_size
    else:
        # ctx len here is usually around 32, so we can use a larger batch size
        # However, we do have backward passes for training the MDBM
        # The divide by 4 shouldn't be necessary, but there's some memory fragmentation issue
        # that causes intermittent OOM errors.
        config.llm_batch_size = (
            activation_collection.LLM_NAME_TO_BATCH_SIZE[config.model_name] // 4
        )

    if args.llm_dtype is not None:
        config.llm_dtype = args.llm_dtype
    else:
        config.llm_dtype = activation_collection.LLM_NAME_TO_DTYPE[config.model_name]

    if args.random_seed is not None:
        config.random_seed = args.random_seed

    if args.train_mdas:
        config.train_mdas = args.train_mdas
        config.num_epochs = 10

    selected_saes = get_saes_from_regex(args.sae_regex_pattern, args.sae_block_pattern)
    assert len(selected_saes) > 0, "No SAEs selected"

    releases = set([release for release, _ in selected_saes])

    print(f"Selected SAEs from releases: {releases}")

    for release, sae in selected_saes:
        print(f"Sample SAEs: {release}, {sae}")

    return config, selected_saes


def arg_parser():
    parser = argparse.ArgumentParser(description="Run RAVEL evaluation")
    parser.add_argument("--random_seed", type=int, default=None, help="Random seed")
    parser.add_argument("--model_name", type=str, required=True, help="Model name")
    parser.add_argument(
        "--sae_regex_pattern",
        type=str,
        required=True,
        help="Regex pattern for SAE selection",
    )
    parser.add_argument(
        "--sae_block_pattern",
        type=str,
        required=True,
        help="Regex pattern for SAE block selection",
    )
    parser.add_argument(
        "--output_folder",
        type=str,
        default="eval_results/ravel",
        help="Output folder",
    )
    parser.add_argument(
        "--force_rerun", action="store_true", help="Force rerun of experiments"
    )
    parser.add_argument(
        "--llm_batch_size",
        type=int,
        default=None,
        help="Batch size for LLM. If None, will be populated using LLM_NAME_TO_BATCH_SIZE",
    )
    parser.add_argument(
        "--llm_dtype",
        type=str,
        default=None,
        choices=[None, "float32", "float64", "float16", "bfloat16"],
        help="Data type for LLM. If None, will be populated using LLM_NAME_TO_DTYPE",
    )
    parser.add_argument(
        "--artifacts_path",
        type=str,
        default="artifacts",
        help="Path to save artifacts",
    )
    parser.add_argument(
        "--train_mdas",
        action="store_true",
        help="Train MDAS instead of SAEs",
    )

    return parser


if __name__ == "__main__":
    """
    python -m sae_bench.evals.ravel.main \
    --sae_regex_pattern "sae_bench_gemma-2-2b_topk_width-2pow14_date-1109" \
    --sae_block_pattern "blocks.12.hook_resid_post__trainer_2" \
    --model_name gemma-2-2b
    """
    args = arg_parser().parse_args()
    device = general_utils.setup_environment()

    start_time = time.time()

    config, selected_saes = create_config_and_selected_saes(args)

    print(selected_saes)

    # create output folder
    os.makedirs(args.output_folder, exist_ok=True)

    # run the evaluation on all selected SAEs
    results_dict = run_eval(
        config,
        selected_saes,
        device,
        args.output_folder,
        args.force_rerun,
        artifacts_path=args.artifacts_path,
    )

    end_time = time.time()

    print(f"Finished evaluation in {end_time - start_time} seconds")


# Use this code snippet to use custom SAE objects
# if __name__ == "__main__":
#     import sae_bench.custom_saes.identity_sae as identity_sae
#     import sae_bench.custom_saes.jumprelu_sae as jumprelu_sae

#     """
#     python evals/ravel/main.py
#     """
#     device = general_utils.setup_environment()

#     start_time = time.time()

#     random_seed = 42
#     output_folder = "eval_results/ravel"

#     model_name = "gemma-2-2b"
#     hook_layer = 20

#     repo_id = "google/gemma-scope-2b-pt-res"
#     filename = f"layer_{hook_layer}/width_16k/average_l0_71/params.npz"
#     sae = jumprelu_sae.load_jumprelu_sae(repo_id, filename, hook_layer)
#     selected_saes = [(f"{repo_id}_{filename}_gemmascope_sae", sae)]

#     config = RAVELEvalConfig(
#         random_seed=random_seed,
#         model_name=model_name,
#     )

#     config.llm_batch_size = activation_collection.LLM_NAME_TO_BATCH_SIZE[config.model_name]
#     config.llm_dtype = activation_collection.LLM_NAME_TO_DTYPE[config.model_name]

#     # create output folder
#     os.makedirs(output_folder, exist_ok=True)

#     # run the evaluation on all selected SAEs
#     results_dict = run_eval(
#         config,
#         selected_saes,
#         device,
#         output_folder,
#         force_rerun=True,
#     )

#     end_time = time.time()

#     print(f"Finished evaluation in {end_time - start_time} seconds")



import gc

import sae_lens
import torch
import torch.nn as nn
import torch.nn.functional as F
from transformers import (
    PreTrainedModel,
    PreTrainedTokenizerBase,
)

import sae_bench.sae_bench_utils.activation_collection as activation_collection
from sae_bench.evals.ravel.eval_config import RAVELEvalConfig
from sae_bench.evals.ravel.mdas import MDAS


class MDBM(nn.Module):
    def __init__(
        self,
        model: PreTrainedModel,
        tokenizer: PreTrainedTokenizerBase,
        config: RAVELEvalConfig,
        sae: sae_lens.SAE,
    ):
        super().__init__()

        self.model = model
        self.tokenizer = tokenizer
        self.sae = sae
        self.layer_intervened = sae.cfg.hook_layer
        self.binary_mask = torch.nn.Parameter(
            torch.zeros(sae.cfg.d_sae, device=model.device, dtype=torch.float32),
            requires_grad=True,
        )
        self.batch_size = config.llm_batch_size
        self.device = model.device
        self.temperature: float = 1

    def create_intervention_hook(
        self,
        source_rep_BD: torch.Tensor,
        base_pos_B: torch.Tensor,
        training_mode: bool = False,
        add_error: bool = False,
    ):
        """
        Creates and returns an intervention hook function that applies a binary mask
        to modify activations.

        Args:
            source_rep_BD: Source representation tensor
            base_pos_B: Base positions tensor
            training_mode: Whether to use sigmoid (training) or hard threshold (eval)
            add_error: Whether to add error to the modified activations - we default to False, as it typically degrades performance

        Returns:
            A hook function that can be registered with a PyTorch module
        """

        def intervention_hook(module, inputs, outputs):
            if isinstance(outputs, tuple):
                resid_BLD = outputs[0]
                rest = outputs[1:]
            else:
                raise ValueError("Unexpected output shape")

            if resid_BLD.shape[1] == 1:
                # This means we are generating with the KV cache and the intervention has already been applied
                return outputs

            with torch.no_grad():
                source_act_BF = self.sae.encode(source_rep_BD)
                resid_BD = resid_BLD[list(range(resid_BLD.shape[0])), base_pos_B, :]
                base_act_BF = self.sae.encode(resid_BD)

            # Use true binary mask in eval mode, sigmoid in training mode
            if not training_mode:
                mask_values_F = (self.binary_mask > 0).to(dtype=self.binary_mask.dtype)
            else:
                mask_values_F = torch.sigmoid(self.binary_mask / self.temperature)

            modified_act_BF = (
                1 - mask_values_F
            ) * base_act_BF + mask_values_F * source_act_BF

            modified_resid_BD = self.sae.decode(
                modified_act_BF.to(dtype=source_rep_BD.dtype)
            )

            if add_error:
                error_BD = resid_BD - self.sae.decode(base_act_BF)
                modified_resid_BD = modified_resid_BD + error_BD

            resid_BLD[list(range(resid_BLD.shape[0])), base_pos_B, :] = (
                modified_resid_BD
            )

            return (resid_BLD, *rest)

        return intervention_hook

    def forward(
        self,
        base_encoding_BL,
        source_encoding_BL,
        base_pos_B,
        source_pos_B,
        training_mode: bool = False,
    ):
        with torch.no_grad():
            # Get source representation
            source_rep_BD = activation_collection.get_layer_activations(
                self.model, self.layer_intervened, source_encoding_BL, source_pos_B
            )

        intervention_hook = self.create_intervention_hook(
            source_rep_BD,
            base_pos_B,
            training_mode,
        )

        handle = activation_collection.get_module(
            self.model, self.layer_intervened
        ).register_forward_hook(intervention_hook)

        logits_BV = self.model(
            input_ids=base_encoding_BL["input_ids"].to(self.model.device),
            attention_mask=base_encoding_BL.get("attention_mask", None),
        ).logits[:, -1, :]

        handle.remove()

        predicted_B = logits_BV.argmax(dim=-1)

        # Format outputs
        predicted_text = []
        for i in range(logits_BV.shape[0]):
            predicted_text.append(self.tokenizer.decode(predicted_B[i]))

        return logits_BV, predicted_text


def compute_loss(intervened_logits_BV, target_attr_B):
    """
    Compute multi-task loss combining:
    - Cause loss: Target attribute should match source
    - Iso loss: Other attributes should match base

    NOTE: For cause loss, target_attr_B is the source attribute value.
    For iso loss, target_attr_B is the base attribute value.
    This is set during dataset creation, so we can just use cross entropy loss with target_attr_B
    for both cause and iso loss.

    Returns:
        Tuple of (loss, accuracy) where accuracy is the raw prediction accuracy
        for the final token
    """
    loss = F.cross_entropy(intervened_logits_BV, target_attr_B)

    # Calculate accuracy
    predictions_B = intervened_logits_BV.argmax(dim=-1)
    accuracy = (predictions_B == target_attr_B).float().mean()

    return loss, accuracy


def get_cause_isolation_score_estimates(
    intervened_logits_BV, source_pred_B, base_pred_B
):
    """
    Calculate cause and isolation scores based on predictions.

    Args:
        intervened_logits_BV: Logits from the intervened model
        source_pred_B: Target predictions from source examples
        base_pred_B: Target predictions from base examples

    Returns:
        Tuple of (cause_score, isolation_score, cause_count, isolation_count)
    """
    predictions_B = intervened_logits_BV.argmax(dim=-1)

    # Identify cause and isolation examples
    is_isolation = base_pred_B == source_pred_B
    is_cause = ~is_isolation

    # Count examples in each category
    cause_count = is_cause.sum().item()
    isolation_count = is_isolation.sum().item()

    # Calculate accuracy for each category
    cause_correct = ((predictions_B == source_pred_B) & is_cause).sum().item()
    isolation_correct = ((predictions_B == base_pred_B) & is_isolation).sum().item()

    # Calculate scores (handle division by zero)
    cause_score = cause_correct / cause_count if cause_count > 0 else 0.0
    isolation_score = (
        isolation_correct / isolation_count if isolation_count > 0 else 0.0
    )

    return cause_score, isolation_score, cause_count, isolation_count


@torch.no_grad()
def get_validation_loss(mdbm: MDBM, val_loader: torch.utils.data.DataLoader):
    """Compute validation loss across the validation dataset"""
    mdbm.eval()
    val_loss = 0
    val_accuracy = 0
    val_batch_count = 0
    val_cause_score = 0
    val_isolation_score = 0
    total_cause_count = 0
    total_isolation_count = 0

    with torch.no_grad():
        for batch in val_loader:
            (
                base_encodings_BL,
                source_encodings_BL,
                base_pos_B,
                source_pos_B,
                base_pred_B,
                source_pred_B,
                base_text_str,
                base_label_str,
            ) = batch
            intervened_logits_BV, _ = mdbm(
                base_encodings_BL, source_encodings_BL, base_pos_B, source_pos_B
            )
            loss, accuracy = compute_loss(intervened_logits_BV, source_pred_B)

            # Calculate cause and isolation scores
            cause_score, isolation_score, cause_count, isolation_count = (
                get_cause_isolation_score_estimates(
                    intervened_logits_BV, source_pred_B, base_pred_B
                )
            )

            val_loss += loss.item()
            val_accuracy += accuracy.item()
            val_cause_score += cause_score * cause_count
            val_isolation_score += isolation_score * isolation_count
            total_cause_count += cause_count
            total_isolation_count += isolation_count
            val_batch_count += 1

    avg_val_loss = val_loss / val_batch_count if val_batch_count > 0 else 0
    avg_val_accuracy = val_accuracy / val_batch_count if val_batch_count > 0 else 0
    avg_val_cause_score = (
        val_cause_score / total_cause_count if total_cause_count > 0 else 0
    )
    avg_val_isolation_score = (
        val_isolation_score / total_isolation_count if total_isolation_count > 0 else 0
    )

    return avg_val_loss, avg_val_accuracy, avg_val_cause_score, avg_val_isolation_score


def train_mdbm(
    model: PreTrainedModel,
    tokenizer: PreTrainedTokenizerBase,
    config: RAVELEvalConfig,
    sae: sae_lens.SAE,
    train_loader,
    val_loader,
    verbose: bool = False,
    train_mdas: bool = False,
) -> MDBM:
    initial_temperature = 1
    final_temperature = 1e-4
    temperature_schedule = torch.logspace(
        torch.log10(torch.tensor(initial_temperature)),
        torch.log10(torch.tensor(final_temperature)),
        config.num_epochs * len(train_loader),
        device=model.device,
        dtype=model.dtype,
    )

    if train_mdas:
        mdbm = MDAS(
            model,
            tokenizer,
            config,
            sae,
        ).to(model.device)
        optimizer = torch.optim.Adam(
            [mdbm.binary_mask, mdbm.transform_matrix], lr=config.learning_rate
        )
        orthonormal = True
    else:
        mdbm = MDBM(
            model,
            tokenizer,
            config,
            sae,
        ).to(model.device)
        optimizer = torch.optim.Adam([mdbm.binary_mask], lr=config.learning_rate)
        orthonormal = False

    if verbose:
        # Get initial validation loss
        (
            initial_val_loss,
            initial_val_accuracy,
            initial_val_cause_score,
            initial_val_isolation_score,
        ) = get_validation_loss(mdbm, val_loader)  # type: ignore
        print(
            f"Initial validation loss: {initial_val_loss:.4f}, accuracy: {initial_val_accuracy:.4f}, cause score: {initial_val_cause_score:.4f}, isolation score: {initial_val_isolation_score:.4f}"
        )

    for epoch in range(config.num_epochs):
        mdbm.train()
        train_loss = 0
        train_accuracy = 0
        batch_count = 0
        log_count = 0

        gc.collect()
        torch.cuda.empty_cache()

        for batch in train_loader:
            mdbm.temperature = temperature_schedule[
                epoch * len(train_loader) + batch_count
            ].item()
            (
                base_encodings_BL,
                source_encodings_BL,
                base_pos_B,
                source_pos_B,
                base_pred_B,
                source_pred_B,
                base_text_str,
                base_label_str,
            ) = batch

            optimizer.zero_grad()

            intervened_logits_BV, _ = mdbm(
                base_encodings_BL,
                source_encodings_BL,
                base_pos_B,
                source_pos_B,
                training_mode=True,
            )
            loss, accuracy = compute_loss(intervened_logits_BV, source_pred_B)

            loss.backward()
            optimizer.step()
            if train_mdas and orthonormal:
                with torch.no_grad():
                    Q, R = torch.linalg.qr(mdbm.transform_matrix, mode="reduced")
                    # Correct sign to enforce det=+1 for rotation matrices
                    det = torch.det(Q)
                    if det < 0:
                        # Flip the sign of one column to make det=+1
                        Q[:, 0] = -Q[:, 0]
                    mdbm.transform_matrix[...] = Q  # type: ignore

            train_loss += loss.item()
            train_accuracy += accuracy.item()
            batch_count += 1
            log_count += 1

            if log_count % 20 == 0 and verbose:
                print(
                    f"Epoch {epoch + 1}/{config.num_epochs} - "
                    f"Train Loss: {train_loss / log_count:.4f}, "
                    f"Train Accuracy: {train_accuracy / log_count:.4f}"
                )
                train_loss = 0
                train_accuracy = 0
                log_count = 0

        avg_train_loss = train_loss / batch_count if batch_count > 0 else 0
        avg_train_accuracy = train_accuracy / batch_count if batch_count > 0 else 0
        # Validation

        # Print losses if verbose
        if verbose:
            (
                avg_val_loss,
                avg_val_accuracy,
                avg_val_cause_score,
                avg_val_isolation_score,
            ) = get_validation_loss(mdbm, val_loader)  # type: ignore
            percent_above_zero = (mdbm.binary_mask > 0).float().mean().item()
            print(
                f"Epoch {epoch + 1}/{config.num_epochs} - "
                f"Train Loss: {avg_train_loss:.4f}, "
                f"Train Accuracy: {avg_train_accuracy:.4f}, "
                f"Val Loss: {avg_val_loss:.4f}, "
                f"Val Accuracy: {avg_val_accuracy:.4f}, "
                f"Percent above zero: {percent_above_zero:.4f}, "
                f"Val Cause Score: {avg_val_cause_score:.4f}, "
                f"Val Isolation Score: {avg_val_isolation_score:.4f}"
            )

        # Early stopping
        # if avg_val_loss < best_val_loss:
        #     best_val_loss = avg_val_loss
        #     patience_counter = 0
        #     if verbose:
        #         print(f"  New best validation loss: {best_val_loss:.4f}")
        # else:
        #     patience_counter += 1
        #     if verbose:
        #         print(f"  No improvement for {patience_counter} epochs")

        # if patience_counter >= config.early_stop_patience:
        #     print(f"Early stopping at epoch {epoch + 1}")
        #     break

    # if verbose:
    #     print(f"Training complete. Best validation loss: {best_val_loss:.4f}")

    return mdbm  # type: ignore



import torch
from jaxtyping import Int
from tqdm import tqdm
from transformers import PreTrainedModel, PreTrainedTokenizerBase


def custom_left_padding(
    tokenizer: PreTrainedTokenizerBase,
    input_ids: list[list[int]],
    pad_to_length: int | None = None,
) -> tuple[
    Int[torch.Tensor, "batch_size seq_len"], Int[torch.Tensor, "batch_size seq_len"]
]:
    """
    Left pad the input ids with the pad token.
    """
    max_length = max(len(ids) for ids in input_ids)
    if pad_to_length is not None:
        max_length = pad_to_length

    if hasattr(tokenizer, "pad_token_id") and tokenizer.pad_token_id is not None:  # type: ignore
        pad_token_id = tokenizer.pad_token_id  # type: ignore
    else:
        pad_token_id = tokenizer.eos_token_id  # type: ignore
    padded_input_ids = [
        [pad_token_id] * (max_length - len(ids)) + ids for ids in input_ids
    ]
    padded_input_ids = torch.tensor(padded_input_ids)
    attention_mask = (padded_input_ids != pad_token_id).long()  # type: ignore
    return padded_input_ids, attention_mask


def generate_batched(
    model: PreTrainedModel,
    tokenizer: PreTrainedTokenizerBase,
    input_ids_BL: Int[torch.Tensor, "batch_size seq_len"] | list[list[int]],
    attention_mask_BL: Int[torch.Tensor, "batch_size seq_len"] | None = None,
    max_new_tokens: int = 8,
    llm_batch_size: int = 32,
    return_first_generated_token: bool = False,
):
    """
    Generate completions for a batch of prompts.
    You can either pass
    1. a tokenized and padded input ids + attention masks as torch tensors
    2. a list of lists of tokenized input ids without padding
    """
    num_total_prompts = len(input_ids_BL)

    generations = []
    for batch_begin in tqdm(
        range(0, num_total_prompts, llm_batch_size),
        desc="Generate completions to test model knowledge",
    ):
        # Draw batch from input_ids_BL
        if isinstance(input_ids_BL, torch.Tensor):
            assert attention_mask_BL is not None, (
                "If input_ids_BL is a torch tensor, attention_mask_BL must also be provided."
            )
            input_ids = input_ids_BL[batch_begin : batch_begin + llm_batch_size].to(
                model.device
            )
            attention_mask = attention_mask_BL[
                batch_begin : batch_begin + llm_batch_size
            ].to(model.device)
        else:
            input_ids, attention_mask = custom_left_padding(
                tokenizer, input_ids_BL[batch_begin : batch_begin + llm_batch_size]
            )
            input_ids = input_ids.to(model.device)
            attention_mask = attention_mask.to(model.device)

        # Generate using huggingface model
        output_ids = model.generate(
            input_ids=input_ids,
            attention_mask=attention_mask,
            max_new_tokens=max_new_tokens,
            do_sample=False,  # greedy decoding for reproducibility
        )  # type: ignore
        generated_ids = output_ids[:, -max_new_tokens:]
        generations.append(generated_ids)

    generations = torch.cat(generations, dim=0)
    generated_strings = tokenizer.batch_decode(generations)
    if return_first_generated_token:
        return generated_strings, generations[:, 0].tolist()
    return generated_strings


if __name__ == "__main__":
    # Test the generation
    from nnsight import LanguageModel  # type: ignore
    from transformers import AutoTokenizer

    device = torch.device("cuda:0")
    model = LanguageModel(
        "eleutherAI/pythia-70m-deduped", device_map=device, dispatch=True
    )
    tokenizer = AutoTokenizer.from_pretrained("eleutherAI/pythia-70m-deduped")
    tokenizer.pad_token = tokenizer.eos_token

    encoded = model.tokenizer.batch_encode_plus(
        ["Hello, world!", "Moin "],
        return_tensors="pt",  # type: ignore
        padding="max_length",  # type: ignore
        max_length=20,  # type: ignore
    ).to(device)
    input_ids_BL = encoded["input_ids"]
    attention_mask_BL = encoded["attention_mask"]

    generated_strings = generate_batched(
        model,  # type: ignore
        tokenizer,
        input_ids_BL,  # type: ignore
        attention_mask_BL,  # type: ignore
        max_new_tokens=10,
    )
    print(generated_strings)



import sae_bench.custom_saes.identity_sae as identity_sae
import sae_bench.evals.autointerp.main as autointerp
import sae_bench.sae_bench_utils.general_utils as general_utils
import sae_bench.sae_bench_utils.sae_selection_utils as sae_selection_utils

if __name__ == "__main__":
    model_name = "gemma-2-2b"
    layers = [5, 12, 19]
    d_model = 2304
    d_mlp = d_model * 4
    llm_dtype = "bfloat16"
    llm_batch_size = 32
    RANDOM_SEED = 42
    device = general_utils.setup_environment()
    force_rerun = False

    with open("openai_api_key.txt") as f:
        api_key = f.read().strip()

    for layer in layers:
        mlp_hook = f"blocks.{layer}.mlp.hook_post"
        mlp_neurons = identity_sae.IdentitySAE(
            model_name, d_mlp, layer, hook_name=mlp_hook
        )  # type: ignore

        resid_hook = f"blocks.{layer}.hook_resid_post"
        resid_dimensions = identity_sae.IdentitySAE(
            model_name, d_model, layer, hook_name=resid_hook
        )  # type: ignore

        selected_saes = [
            (f"identity_mlp_{model_name}_layer_{layer}", mlp_neurons),
            (f"identity_resid_{model_name}_layer_{layer}", resid_dimensions),
        ]

        autointerp.run_eval(
            autointerp.AutoInterpEvalConfig(
                model_name=model_name,
                random_seed=RANDOM_SEED,
                llm_batch_size=llm_batch_size,
                llm_dtype=llm_dtype,
            ),
            selected_saes,  # type: ignore
            device,
            api_key,
            "eval_results/autointerp",
            force_rerun,
        )

    for layer in layers:
        sae_regex_patterns = [
            r"gemma-scope-2b-pt-mlp-canonical",
            r"gemma-scope-2b-pt-res-canonical",
        ]
        sae_block_patterns = [
            rf".*layer_{layer}.*(16k).*",
            rf".*layer_{layer}.*(16k).*",
        ]

        selected_saes = sae_selection_utils.select_saes_multiple_patterns(
            sae_regex_patterns, sae_block_patterns
        )

        autointerp.run_eval(
            autointerp.AutoInterpEvalConfig(
                model_name=model_name,
                random_seed=RANDOM_SEED,
                llm_batch_size=llm_batch_size,
                llm_dtype=llm_dtype,
            ),
            selected_saes,
            device,
            api_key,
            "eval_results/autointerp",
            force_rerun,
        )



import os
from typing import Any

import torch
from tqdm import tqdm

import sae_bench.evals.absorption.main as absorption
import sae_bench.evals.autointerp.main as autointerp
import sae_bench.evals.core.main as core
import sae_bench.evals.ravel.main as ravel
import sae_bench.evals.scr_and_tpp.main as scr_and_tpp
import sae_bench.evals.sparse_probing.main as sparse_probing
import sae_bench.evals.unlearning.main as unlearning
import sae_bench.sae_bench_utils.general_utils as general_utils

RANDOM_SEED = 42

MODEL_CONFIGS = {
    "pythia-70m-deduped": {
        "batch_size": 512,
        "dtype": "float32",
        "layers": [3, 4],
        "d_model": 512,
    },
    "gemma-2-2b": {
        "batch_size": 32,
        "dtype": "bfloat16",
        "layers": [12],
        "d_model": 2304,
    },
}

output_folders = {
    "absorption": "eval_results/absorption",
    "autointerp": "eval_results/autointerp",
    "core": "eval_results/core",
    "scr": "eval_results/scr",
    "tpp": "eval_results/tpp",
    "sparse_probing": "eval_results/sparse_probing",
    "unlearning": "eval_results/unlearning",
    "ravel": "eval_results/ravel",
}


def run_evals(
    model_name: str,
    selected_saes: list[tuple[str, Any]],
    llm_batch_size: int,
    llm_dtype: str,
    device: str,
    eval_types: list[str],
    api_key: str | None = None,
    force_rerun: bool = False,
    save_activations: bool = False,
):
    """Run selected evaluations for the given model and SAEs."""

    if model_name not in MODEL_CONFIGS:
        raise ValueError(f"Unsupported model: {model_name}")

    # Mapping of eval types to their functions and output paths
    eval_runners = {
        "absorption": (
            lambda: absorption.run_eval(
                absorption.AbsorptionEvalConfig(
                    model_name=model_name,
                    random_seed=RANDOM_SEED,
                    llm_batch_size=llm_batch_size,
                    llm_dtype=llm_dtype,
                ),
                selected_saes,
                device,
                "eval_results/absorption",
                force_rerun,
            )
        ),
        "autointerp": (
            lambda: autointerp.run_eval(
                autointerp.AutoInterpEvalConfig(
                    model_name=model_name,
                    random_seed=RANDOM_SEED,
                    llm_batch_size=llm_batch_size,
                    llm_dtype=llm_dtype,
                ),
                selected_saes,
                device,
                api_key,  # type: ignore
                "eval_results/autointerp",
                force_rerun,
            )
        ),
        # TODO: Do a better job of setting num_batches and batch size
        "core": (
            lambda: core.multiple_evals(
                selected_saes=selected_saes,
                n_eval_reconstruction_batches=200,
                n_eval_sparsity_variance_batches=2000,
                eval_batch_size_prompts=16,
                compute_featurewise_density_statistics=True,
                compute_featurewise_weight_based_metrics=True,
                exclude_special_tokens_from_reconstruction=True,
                dataset="Skylion007/openwebtext",
                context_size=128,
                output_folder="eval_results/core",
                verbose=True,
                dtype=llm_dtype,
                device=device,
            )
        ),
        "ravel": (
            lambda: ravel.run_eval(
                ravel.RAVELEvalConfig(
                    model_name=model_name,
                    random_seed=RANDOM_SEED,
                    llm_batch_size=llm_batch_size // 4,
                    llm_dtype=llm_dtype,
                ),
                selected_saes,
                device,
                "eval_results/ravel",
                force_rerun,
            )
        ),
        "scr": (
            lambda: scr_and_tpp.run_eval(
                scr_and_tpp.ScrAndTppEvalConfig(
                    model_name=model_name,
                    random_seed=RANDOM_SEED,
                    perform_scr=True,
                    llm_batch_size=llm_batch_size,
                    llm_dtype=llm_dtype,
                ),
                selected_saes,
                device,
                "eval_results",  # We add scr or tpp depending on perform_scr
                force_rerun,
                clean_up_activations=True,
                save_activations=save_activations,
            )
        ),
        "tpp": (
            lambda: scr_and_tpp.run_eval(
                scr_and_tpp.ScrAndTppEvalConfig(
                    model_name=model_name,
                    random_seed=RANDOM_SEED,
                    perform_scr=False,
                    llm_batch_size=llm_batch_size,
                    llm_dtype=llm_dtype,
                ),
                selected_saes,
                device,
                "eval_results",  # We add scr or tpp depending on perform_scr
                force_rerun,
                clean_up_activations=True,
                save_activations=save_activations,
            )
        ),
        "sparse_probing": (
            lambda: sparse_probing.run_eval(
                sparse_probing.SparseProbingEvalConfig(
                    model_name=model_name,
                    random_seed=RANDOM_SEED,
                    llm_batch_size=llm_batch_size,
                    llm_dtype=llm_dtype,
                ),
                selected_saes,
                device,
                "eval_results/sparse_probing",
                force_rerun,
                clean_up_activations=True,
                save_activations=save_activations,
            )
        ),
        "unlearning": (
            lambda: unlearning.run_eval(
                unlearning.UnlearningEvalConfig(
                    model_name="gemma-2-2b-it",
                    random_seed=RANDOM_SEED,
                    llm_dtype=llm_dtype,
                    llm_batch_size=llm_batch_size // 8,
                ),
                selected_saes,
                device,
                "eval_results/unlearning",
                force_rerun,
            )
        ),
    }

    # Run selected evaluations
    for eval_type in tqdm(eval_types, desc="Evaluations"):
        if eval_type == "autointerp" and api_key is None:
            print("Skipping autointerp evaluation due to missing API key")
            continue
        if eval_type == "unlearning":
            if model_name != "gemma-2-2b":
                print("Skipping unlearning evaluation for non-GEMMA model")
                continue
            print("Skipping, need to clean up unlearning interface")
            continue  # TODO:
            if not os.path.exists(
                "./sae_bench/evals/unlearning/data/bio-forget-corpus.jsonl"
            ):
                print(
                    "Skipping unlearning evaluation due to missing bio-forget-corpus.jsonl"
                )
                continue

        print(f"\n\n\nRunning {eval_type} evaluation\n\n\n")

        if eval_type in eval_runners:
            os.makedirs(output_folders[eval_type], exist_ok=True)
            eval_runners[eval_type]()


if __name__ == "__main__":
    import sae_bench.custom_saes.identity_sae as identity_sae

    device = general_utils.setup_environment()

    model_name = "pythia-70m-deduped"
    model_name = "gemma-2-2b"
    d_model = MODEL_CONFIGS[model_name]["d_model"]
    llm_batch_size = MODEL_CONFIGS[model_name]["batch_size"]
    llm_dtype = MODEL_CONFIGS[model_name]["dtype"]

    # Note: Unlearning is not recommended for models with < 2B parameters and we recommend an instruct tuned model
    # Unlearning will also require requesting permission for the WMDP dataset (see unlearning/README.md)
    # Absorption not recommended for models < 2B parameters

    # Select your eval types here.
    eval_types = [
        "absorption",
        "autointerp",
        "core",
        "ravel",
        "scr",
        "tpp",
        "sparse_probing",
        "unlearning",
    ]

    if "autointerp" in eval_types:
        try:
            with open("openai_api_key.txt") as f:
                api_key = f.read().strip()
        except FileNotFoundError:
            raise Exception("Please create openai_api_key.txt with your API key")
    else:
        api_key = None

    if "unlearning" in eval_types:
        if not os.path.exists(
            "./sae_bench/evals/unlearning/data/bio-forget-corpus.jsonl"
        ):
            raise Exception(
                "Please download bio-forget-corpus.jsonl for unlearning evaluation"
            )

    # If evaluating multiple SAEs on the same layer, set save_activations to True
    # This will require at least 100GB of disk space
    save_activations = False

    for hook_layer in MODEL_CONFIGS[model_name]["layers"]:
        sae = identity_sae.IdentitySAE(
            d_model,
            model_name,
            hook_layer,
            device=torch.device(device),
            dtype=general_utils.str_to_dtype(llm_dtype),
        )  # type: ignore
        selected_saes = [(f"{model_name}_layer_{hook_layer}_identity_sae", sae)]

        # This will evaluate PCA SAEs
        # sae = pca_sae.PCASAE(
        #     d_model,
        #     model_name,
        #     hook_layer,
        #     device=torch.device(device),
        #     dtype=general_utils.str_to_dtype(llm_dtype),
        # )
        # filename = f"pca_gemma-2-2b_blocks.{hook_layer}.hook_resid_post.pt"
        # sae.load_from_file(filename)
        # selected_saes = [(f"{model_name}_layer_{hook_layer}_pca_sae", sae)]

        for sae_name, sae in selected_saes:
            sae = sae.to(dtype=general_utils.str_to_dtype(llm_dtype))
            sae.cfg.dtype = llm_dtype

        run_evals(
            model_name,
            selected_saes,
            llm_batch_size,
            llm_dtype,
            device,
            eval_types=eval_types,
            api_key=api_key,
            force_rerun=False,
            save_activations=False,
        )



from dataclasses import dataclass
from pathlib import Path

import pandas as pd
from sae_lens import SAE
from tqdm.autonotebook import tqdm
from transformer_lens import HookedTransformer

from sae_bench.evals.absorption.common import (
    PROBES_DIR,
    RESULTS_DIR,
    get_or_make_dir,
    load_df_or_run,
    load_experiment_df,
    load_probe,
)
from sae_bench.evals.absorption.feature_absorption_calculator import (
    FeatureAbsorptionCalculator,
)
from sae_bench.evals.absorption.k_sparse_probing import (
    SPARSE_PROBING_EXPERIMENT_NAME,
    add_feature_splits_to_metrics_df,
    get_sparse_probing_metrics_filename,
    get_sparse_probing_raw_results_filename,
)
from sae_bench.evals.absorption.probing import LinearProbe
from sae_bench.evals.absorption.prompting import (
    first_letter_formatter,
)
from sae_bench.evals.absorption.vocab import LETTERS, get_alpha_tokens

FEATURE_ABSORPTION_EXPERIMENT_NAME = "feature_absorption"

# the cosine similarity between the top projecting feature and the probe must be at least this high (full absorption only)
FULL_ABSORPTION_PROBE_COS_THRESHOLD = 0.025

# the cosine similarity between potential absorbing latents and the probe must be at least this high (absorption fraction only)
ABSORPTION_FRACTION_PROBE_COS_THRESHOLD = 0.1

# the total probe projection of the potential absorbing latents must contribute at least this much to the probe projection to count as absorption (both absorption metrics)
ABSORPTION_PROBE_PROJECTION_PROPORTION_THRESHOLD = 0.4

# the maximum number of latents that can be considered to collectively compensate for the reduced activation of a potentially absorbed latent (absorption fraction only)
ABSORPTION_FRACTION_MAX_ABSORBING_LATENTS = 3


@dataclass
class StatsAndLikelyFalseNegativeResults:
    probe_true_positives: int
    split_feats_true_positives: int
    split_feats: list[int]
    potential_false_negatives: list[str]


def calculate_projection_and_cos_sims(
    calculator: FeatureAbsorptionCalculator,
    sae: SAE,
    probe: LinearProbe,
    layer: int,
    likely_negs: dict[str, StatsAndLikelyFalseNegativeResults],
) -> pd.DataFrame:
    results = []
    for letter, stats in tqdm(likely_negs.items()):
        assert calculator.model.tokenizer is not None
        absorption_results = calculator.calculate_absorption(
            sae,
            layer=layer,
            words=stats.potential_false_negatives,
            probe_direction=probe.weights[LETTERS.index(letter)],
            main_feature_ids=stats.split_feats,
            show_progress=False,
        )
        for sample in absorption_results.word_results:
            top_feat_score = sample.top_projection_feature_scores[0]
            second_feat_score = sample.top_projection_feature_scores[1]
            result = {
                "letter": letter,
                "token": sample.word,
                "prompt": sample.prompt,
                "num_probe_true_positives": stats.probe_true_positives,
                "split_feats": stats.split_feats,
                "split_feat_acts": [
                    score.activation for score in sample.main_feature_scores
                ],
                "split_feat_probe_cos": [
                    score.probe_cos_sim for score in sample.main_feature_scores
                ],
                "top_projection_feat": top_feat_score.feature_id,
                "top_probe_projection": top_feat_score.probe_projection,
                "top_projection_feat_probe_cos": top_feat_score.probe_cos_sim,
                "second_projection_feat": second_feat_score.feature_id,
                "second_probe_projection": second_feat_score.probe_projection,
                "second_projection_feat_probe_cos": second_feat_score.probe_cos_sim,
                "probe_projections": [
                    score.probe_projection
                    for score in sample.top_projection_feature_scores
                ],
                "projection_feats": [
                    score.feature_id for score in sample.top_projection_feature_scores
                ],
                "projection_feat_acts": [
                    score.activation for score in sample.top_projection_feature_scores
                ],
                "projection_feat_probe_cos": [
                    score.probe_cos_sim
                    for score in sample.top_projection_feature_scores
                ],
                "absorption_fraction": sample.absorption_fraction,
                "is_full_absorption": sample.is_full_absorption,
            }
            results.append(result)
    result_df = pd.DataFrame(results)
    return result_df


def get_stats_and_likely_false_negative_tokens(
    metrics_df: pd.DataFrame,
    sae_name: str,
    layer: int,
    sparse_probing_task_output_dir: Path,
) -> dict[str, StatsAndLikelyFalseNegativeResults]:
    """
    Examine the k-sparse probing results and look for false-negative cases where the k top feats don't fire but our LR probe does
    """
    results: dict[str, StatsAndLikelyFalseNegativeResults] = {}
    raw_df = load_experiment_df(
        SPARSE_PROBING_EXPERIMENT_NAME,
        sparse_probing_task_output_dir
        / get_sparse_probing_raw_results_filename(sae_name, layer),
    )
    for letter in LETTERS:
        split_feats = metrics_df[metrics_df["letter"] == letter]["split_feats"].iloc(  # type: ignore
            0
        )[0]
        k = len(split_feats)
        potential_false_negatives = raw_df[
            (raw_df["answer_letter"] == letter) & (raw_df[f"score_probe_{letter}"] > 0)
        ]["token"].tolist()
        num_split_feats_true_positives = raw_df[
            (raw_df["answer_letter"] == letter)
            & (raw_df[f"score_probe_{letter}"] > 0)
            & (raw_df[f"score_sparse_sae_{letter}_k_{k}"] > 0)
        ].shape[0]
        num_probe_true_positives = raw_df[
            (raw_df["answer_letter"] == letter) & (raw_df[f"score_probe_{letter}"] > 0)
        ].shape[0]
        results[letter] = StatsAndLikelyFalseNegativeResults(
            probe_true_positives=num_probe_true_positives,
            split_feats_true_positives=num_split_feats_true_positives,
            split_feats=split_feats,
            potential_false_negatives=potential_false_negatives,
        )
    return results


def load_and_run_calculate_projections_and_cos_sims(
    model: HookedTransformer,
    sae: SAE,
    calculator: FeatureAbsorptionCalculator,
    metrics_df: pd.DataFrame,
    sae_name: str,
    layer: int,
    probes_dir: Path | str,
    sparse_probing_task_output_dir: Path,
    device: str,
) -> pd.DataFrame:
    probe = load_probe(
        model_name=model.cfg.model_name,
        layer=layer,
        probes_dir=probes_dir,
        device=device,
    )
    likely_negs = get_stats_and_likely_false_negative_tokens(
        metrics_df, sae_name, layer, sparse_probing_task_output_dir
    )
    return calculate_projection_and_cos_sims(
        calculator, sae, probe, likely_negs=likely_negs, layer=layer
    )


def run_feature_absortion_experiment(
    model: HookedTransformer,
    sae: SAE,
    layer: int,
    sae_name: str,
    max_k_value: int,
    prompt_template: str,
    prompt_token_pos: int,
    device: str,
    experiment_dir: Path | str = RESULTS_DIR / FEATURE_ABSORPTION_EXPERIMENT_NAME,
    sparse_probing_experiment_dir: Path | str = RESULTS_DIR
    / SPARSE_PROBING_EXPERIMENT_NAME,
    probes_dir: Path | str = PROBES_DIR,
    force: bool = False,
    feature_split_f1_jump_threshold: float = 0.03,
    batch_size: int = 10,
) -> pd.DataFrame:
    """
    NOTE: this experiments requires the results of the k-sparse probing experiments. Make sure to run them first.
    """
    task_output_dir = get_or_make_dir(experiment_dir) / sae_name
    sparse_probing_task_output_dir = (
        get_or_make_dir(sparse_probing_experiment_dir) / sae_name
    )

    vocab = get_alpha_tokens(model.tokenizer)  # type: ignore
    calculator = FeatureAbsorptionCalculator(
        model=model,
        icl_word_list=vocab,
        max_icl_examples=10,
        base_template=prompt_template,
        answer_formatter=first_letter_formatter(),
        word_token_pos=prompt_token_pos,
        full_absorption_probe_cos_sim_threshold=FULL_ABSORPTION_PROBE_COS_THRESHOLD,
        absorption_fraction_probe_cos_sim_threshold=ABSORPTION_FRACTION_PROBE_COS_THRESHOLD,
        probe_projection_proportion_threshold=ABSORPTION_PROBE_PROJECTION_PROPORTION_THRESHOLD,
        absorption_fraction_max_absorbing_latents=ABSORPTION_FRACTION_MAX_ABSORBING_LATENTS,
        batch_size=batch_size,
    )
    metrics_df = load_experiment_df(
        SPARSE_PROBING_EXPERIMENT_NAME,
        sparse_probing_task_output_dir
        / get_sparse_probing_metrics_filename(sae_name, layer),
    )
    add_feature_splits_to_metrics_df(
        metrics_df,
        max_k_value=max_k_value,
        f1_jump_threshold=feature_split_f1_jump_threshold,
    )
    df_path = task_output_dir / f"layer_{layer}_{sae_name}.parquet"
    df = load_df_or_run(
        lambda: load_and_run_calculate_projections_and_cos_sims(
            model,
            sae,
            calculator,
            metrics_df,
            sae_name=sae_name,
            layer=layer,
            probes_dir=probes_dir,
            sparse_probing_task_output_dir=sparse_probing_task_output_dir,
            device=device,
        ),
        df_path,
        force=force,
    )
    return df
